/*
 * Created on Jul 2, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine;

import java.awt.Color;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;

import org.apache.batik.transcoder.TranscoderInput;
import org.apache.batik.transcoder.TranscoderOutput;
import org.apache.batik.transcoder.image.ImageTranscoder;
import org.apache.batik.transcoder.image.JPEGTranscoder;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.apache.batik.transcoder.image.TIFFTranscoder;
import org.apache.fop.svg.PDFTranscoder;

import diagnose.graphicengine.devices.SVGExportDevice;
import diagnose.graphicengine.elements.Document;
import diagnose.utils.Lang;
import diagnose.utils.Log;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.serialization.XmlConstants;

/**
 * Export to other formats
 */
public class Export {
    
    public static final int SVG     = 1;
    public static final int PNG     = 2;
    public static final int TIFF    = 3;
    public static final int PDF     = 4;
    public static final int JPEG    = 5;
    
    public static void export(File file, Document document, int format) throws Exception {
        
        
        // all exported coordinates have to have positive values
        document = (Document)document.clone();
        Rect rect = document.getRectBounds();
        Point direction = new Point(0, 0);
        direction.x = -rect.getPosition().x;
        direction.y = -rect.getPosition().y;
        document.move(direction);
        
        // export
        rect = document.getRectBounds();
        double width = rect.getWidth();
        double height = rect.getHeight();
        
        ImageTranscoder trans;
        
        switch (format) {
        
        case SVG:
            exportToSVG(document, new FileWriter(file), width, height);
            break;
            
        case PNG:
            trans = new PNGTranscoder();
            exportUsingBatik(file, document, trans, width, height);
            break;
            
        case TIFF:
            trans = new TIFFTranscoder();
            trans.addTranscodingHint(TIFFTranscoder.KEY_BACKGROUND_COLOR, new Color(255, 255, 255));
            exportUsingBatik(file, document, trans, width, height);
            break;
            
        case JPEG:
            trans = new JPEGTranscoder();
            trans.addTranscodingHint(JPEGTranscoder.KEY_QUALITY, new Float(0.9));
            exportUsingBatik(file, document, trans, width, height);
            break;
            
        case PDF:
            exportPDF(file, document, width, height);
            break;
            
        default:
            Log.warn(Lang.get("log.unknownexportformat"));
            break;
        }
    }
    
    /**
     * Export files using batik transcoder
     */
    private static void exportUsingBatik(File file, Document document, 
                            ImageTranscoder transcoder, double width, double height) 
                                throws Exception {
        
        StringWriter swr = new StringWriter();
        exportToSVG(document, swr, width, height);
        StringReader srd = new StringReader(swr.toString());
    
        TranscoderInput input = new TranscoderInput(srd);
        
        OutputStream ostream = new FileOutputStream(file);
        
        TranscoderOutput output = new TranscoderOutput(ostream);
        transcoder.transcode(input, output);
        ostream.flush();
        ostream.close();
    }
    
    /**
     * Export files using pdf transcoder
     */
    private static void exportPDF(File file, Document document, double width, double height) throws Exception {
        
        StringWriter swr = new StringWriter();
        exportToSVG(document, swr, width, height);
        StringReader srd = new StringReader(swr.toString());
        
        PDFTranscoder transcoder = new PDFTranscoder();
    
        TranscoderInput input = new TranscoderInput(srd);
        
        OutputStream ostream = new FileOutputStream(file);
        TranscoderOutput output = new TranscoderOutput(ostream);
        transcoder.transcode(input, output);
        ostream.flush();
        ostream.close();
    }
    
    
    /**
     * Writes document into specified writer in SVG form
     */
    private static void exportToSVG(Document document, Writer wr, double width, double height) throws IOException {
        SVGExportDevice exporter = new SVGExportDevice(wr, width, height);
        Object oldVal = document.getProperties().getProperty(XmlConstants.PROP_SHOW_GUIDELINES).getValue();
        document.getProperties().getProperty(XmlConstants.PROP_SHOW_GUIDELINES).setValue(new Boolean(false));
        document.paint(exporter);
        document.getProperties().getProperty(XmlConstants.PROP_SHOW_GUIDELINES).setValue(oldVal);
        exporter.close();
        wr.flush();
        wr.close();
    }
}