/*
 * Created on Feb 6, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine;

import java.awt.Color;
import java.awt.Cursor;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import org.jdom.JDOMException;

import diagnose.graphicengine.contexts.DefaultContext;
import diagnose.graphicengine.contexts.ElementCreationContext;
import diagnose.graphicengine.contexts.EventContext;
import diagnose.graphicengine.contexts.PaddingContext;
import diagnose.graphicengine.contexts.ZoomContext;
import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.Document;
import diagnose.graphicengine.elements.LibraryElement;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.graphicengine.elements.system.GuideLine;
import diagnose.graphicengine.elements.system.Selection;
import diagnose.graphicengine.events.AddElementEvent;
import diagnose.graphicengine.events.AlignEvent;
import diagnose.graphicengine.events.GroupSelectionEvent;
import diagnose.graphicengine.events.MoveElementEvent;
import diagnose.graphicengine.events.PasteSelectionEvent;
import diagnose.graphicengine.events.PersistentEvent;
import diagnose.graphicengine.events.PropertiesChangeEvent;
import diagnose.graphicengine.events.RemoveSelectionEvent;
import diagnose.graphicengine.events.SendBackEvent;
import diagnose.graphicengine.events.SendFrontEvent;
import diagnose.graphicengine.events.SplitSelectionEvent;
import diagnose.graphicengine.libraries.DependencyException;
import diagnose.graphicengine.libraries.Librarian;
import diagnose.graphicengine.libraries.Library;
import diagnose.graphicengine.relation.RelationManager;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.CursorFactory;
import diagnose.utils.DefaultArrays;
import diagnose.utils.Lang;
import diagnose.utils.Log;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.BoolProperty;
import diagnose.utils.properties.ColorProperty;
import diagnose.utils.properties.DoubleProperty;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.Property;
import diagnose.utils.properties.PropertyChangeList;
import diagnose.utils.properties.PropertyChangeLog;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.properties.SelectPropertyDescription;
import diagnose.utils.properties.StringProperty;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Serializer;
import diagnose.utils.serialization.XmlConstants;

/**
 * Manager - gate between GUI and application logic
 * Contains many useful informations for other classes
 */
public class Manager {
    
    public static final double MINIMAL_PADDING_X = -1000;
    public static final double MINIMAL_PADDING_Y = -1000;
    public static final double MAXIMAL_PADDING_X = 1000;
    public static final double MAXIMAL_PADDING_Y = 1000;
    public static final double MIN_ZOOM = 0.2;
    public static final double MAX_ZOOM = 16;
    
    public static final double PRECISSNESS = 0.5;
    
    private GraphicArea parent;
    
    private Document document;
    private Selection selection = new Selection();
    private static Selection clipboard;
    private UndoRedoStack undoRedoStack = new UndoRedoStack();
    private int snapPrecissness = 3;
    
    private Librarian librarian;
    private RelationManager relationManager;
    private EventContext activeContext = null;
    
    private Properties defaultProperties = new Properties();
    
    private UIInvoker propertyDialogInvoker = null;
    private UIInvoker menuRefreshInvoker = null;
    private UIInvoker contextMenuInvoker = null;
    private UIInvoker viewChangeInvoker = null;
    private UIInvoker mouseChangeInvoker = null;
    private UIInvoker defaultPropertyChangeInvoker = null;
    private UIInvoker notFoundElementsInvoker = null;
    
    private String documentPath = null;
    private boolean documentChanged = true;
    
    public static final String LIB_CONFIG_PATH = "libconfig.xml";
    
    public static PropertyDescriptor gridPropertyDescriptor = new PropertyDescriptor();
    
    static {
        CursorFactory.createCursor("magnif_cur", new java.awt.Point(8, 8));
        CursorFactory.createCursor("hand_cur", new java.awt.Point(8, 8));
        
        gridPropertyDescriptor.setDescription(XmlConstants.PROP_GRID_SIZE, Lang.get("property.gridsize"), Lang.get("category.grid"), 20);
        gridPropertyDescriptor.setDescription(XmlConstants.PROP_SHOW_GRID, Lang.get("property.showgrid"), Lang.get("category.grid"), 21);
        gridPropertyDescriptor.setDescription(XmlConstants.PROP_SNAP_TO_GRID, Lang.get("property.snaptogrid"), Lang.get("category.grid"), 22);
    }
    
    public Manager(GraphicArea parent) {
        super();
        setDocument(new Document());
        this.parent = parent;
        
        this.activeContext = new DefaultContext(this, parent.getDevice());
        parent.getDevice().setEventContext(activeContext);
        
        setZoom(1);
        setPadding(new Point(0, 0));
        
        try {
            librarian = DefaultDeserializer.deserializeLibrarian(new File(LIB_CONFIG_PATH));
        } catch (Exception e) {
            librarian = new Librarian();
            Log.warn(Lang.get("log.unabletoloadlibconfig"));
        }
        
        defaultProperties.setProperty(new ColorProperty(XmlConstants.PROP_COLOR, Color.BLACK));
        defaultProperties.setProperty(new ColorProperty(XmlConstants.PROP_BACKGROUND, Color.WHITE));
        defaultProperties.setProperty(new DoubleProperty(XmlConstants.PROP_LINE_WIDTH, 0.35));
        defaultProperties.setProperty(new IntProperty(XmlConstants.PROP_LINE_STYLE, AbstractDevice.CLEAR_LINE));
        defaultProperties.setProperty(new StringProperty(XmlConstants.PROP_FONT_FACE, Lang.get("application.defaultfont")));
        defaultProperties.setProperty(new IntProperty(XmlConstants.PROP_ALIGN, AbstractDevice.ALIGN_LEFT));
        defaultProperties.setProperty(new IntProperty(XmlConstants.PROP_FONT_SIZE, 12));
        defaultProperties.setProperty(new BoolProperty(XmlConstants.PROP_BOLD, false));
        defaultProperties.setProperty(new BoolProperty(XmlConstants.PROP_ITALIC, false));
        defaultProperties.setProperty(new BoolProperty(XmlConstants.PROP_UNDERLINE, false));
    }
    
    public Document getDocument() {
        return document;
    }
    
    protected void setDocument(Document document) {
        this.document = document;
        relationManager = new RelationManager(document);
        relationManager.updateRelations();
        repaint();
    }    
 
    public double getGridSize() {
        return getDocument().getProperties().getDoubleValue(XmlConstants.PROP_GRID_SIZE);
    }
    
    public void setGridSize(int gridSize) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_GRID_SIZE).setValue(new Integer(gridSize));
    }
    
    public boolean isSnapToGrid() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SNAP_TO_GRID);
    }
    
    public void setSnapToGrid(boolean snapToGrid) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SNAP_TO_GRID).setValue(new Boolean(snapToGrid));
        invokeMenuRefresh();
    }
    
    public boolean isSnapToGuideLines() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SNAP_TO_GUIDELINES);
    }
    
    public void setSnapToGuideLines(boolean snapToGuideLines) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SNAP_TO_GUIDELINES).setValue(new Boolean(snapToGuideLines));
        invokeMenuRefresh();
    }
    
    public boolean isSnapToElements() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SNAP_TO_OTHER_ELEMENT);
    }
    
    public void setSnapToElements(boolean snapToElements) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SNAP_TO_OTHER_ELEMENT).setValue(new Boolean(snapToElements));
        invokeMenuRefresh();
    }
    
    public boolean isSnapTech() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SNAP_TECH);
    }
    
    public void setSnapTech(boolean snapTech) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SNAP_TECH).setValue(new Boolean(snapTech));
        refreshDocumentRelations();
        invokeMenuRefresh();
    }
    
    public void setShowGrid(boolean showGrid) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SHOW_GRID).setValue(new Boolean(showGrid));
        invokeMenuRefresh();
        repaint();
    }
    
    public boolean isShowGrid() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SHOW_GRID);
    }
    
    public void setShowGuideLines(boolean showGuideLines) {
        getDocument().getProperties().getProperty(XmlConstants.PROP_SHOW_GUIDELINES).setValue(new Boolean(showGuideLines));
        invokeMenuRefresh();
        repaint();
    }
    
    public boolean isShowGuideLines() {
        return getDocument().getProperties().getBooleanValue(XmlConstants.PROP_SHOW_GUIDELINES);
    }
    
    public int getSnapPrecissness() {
        return snapPrecissness;
    }
    
    public void setSnapPrecissness(int snapPrecissness) {
        this.snapPrecissness = snapPrecissness;
    }
    
    protected void setActiveContext(EventContext activeContext) {
        if (activeContext != null) activeContext.cancelContext();
        parent.getDevice().setEventContext(activeContext);
        this.activeContext = activeContext;
        this.activeContext.activateContext();
        invokeViewChange();
    }
    
    public double getZoom() {
        return getDocument().getProperties().getDoubleValue(XmlConstants.PROP_ZOOM);
    }
    
    public void setZoom(double zoom) {
        double oldZoom = getZoom();
        if (zoom < MIN_ZOOM) zoom = MIN_ZOOM;
        if (zoom > MAX_ZOOM) zoom = MAX_ZOOM;
        Point newPadding = (Point)getPadding().clone();
        newPadding.multiply(zoom / oldZoom);
        
        getDocument().getProperties().getProperty(XmlConstants.PROP_ZOOM).setValue(new Double(zoom));
        setPadding(newPadding);
        repaint();
    }
    
    public Point getPadding() {
        return getDocument().getProperties().getPointValue(XmlConstants.PROP_PADDING);
    }
    
    public void setPadding(Point padding) {
        if (padding.x < MINIMAL_PADDING_X * getZoom()) padding.x = MINIMAL_PADDING_X * getZoom();
        if (padding.y < MINIMAL_PADDING_Y * getZoom()) padding.y = MINIMAL_PADDING_Y * getZoom();
        if (padding.x > MAXIMAL_PADDING_X * getZoom()) padding.x = MAXIMAL_PADDING_X * getZoom();
        if (padding.y > MAXIMAL_PADDING_Y * getZoom()) padding.y = MAXIMAL_PADDING_Y * getZoom();
        getDocument().getProperties().getProperty(XmlConstants.PROP_PADDING).setValue(padding);
        repaint();
    }
    
    public void setCursor(Cursor cur) {
        parent.setCursor(cur);
    }
    
    protected Selection getClipboard() {
        return clipboard;
    }
    
    protected void setClipboard(Selection clipboard) {
        Manager.clipboard = clipboard;
    }
    
    /**
     * Invokes painting on level of elements
     */
    protected void paint() {

        parent.getDevice().setZoom(getZoom());
        parent.getDevice().setPadding(getPadding());
        getDocument().paint(parent.getDevice());
        selection.getControlPoints().paint(parent.getDevice());
        activeContext.paint(parent.getDevice());
    }
    
    /**
     * Calls for repaiting whole component
     */
    public void repaint() {
        if (parent != null) parent.repaint();
    }
        
    /**
     * Drags specified element in to specified direction
     */
    public void moveElement(VisualElement elt, Point direction) {
        performEvent(new MoveElementEvent(this, direction, elt));
        repaint();
    }   
    
    /**
     * This function lets create new object in document
     */
    public void setupCreatingObject(CommonElement elt) {
        setActiveContext(new ElementCreationContext(this, parent.getDevice(), elt));
    }
  
    /**
     * Performs persistant event
     */
    protected void performEvent(PersistentEvent event) {
        event.perform();
        documentChanged = true;
        undoRedoStack.registerEvent(event);
        refreshDocumentRelations();
    }
    
    /**
     * Adds new element into the document
     */
    public void addElement(CommonElement newElement) {
        performEvent(new AddElementEvent(this, newElement));
        if (newElement instanceof SubmissiveElement) 
            relationManager.updateRelations((SubmissiveElement)newElement);
    }
    
    /**
     * Removes all elements from selection
     */
    public Selection clearSelectionEvent() {
        selection = new Selection();
        repaint();
        menuRefreshInvoker.invoke(this);
        return selection;
    }
    
    /**
     * Adds element into selection
     */
    public Selection addElementToSelectionEvent(CommonElement elt) {
        if (selection.hasElement(elt)) return selection;
        selection.addElement(elt);
        menuRefreshInvoker.invoke(this);
        refreshDocumentRelations();
        return selection;
    }
    
    /**
     * Creates selection from objects in rectangular area
     */
    public Selection setSelectionByRect(Rect r, boolean leaveOldSelection) {
        if (!leaveOldSelection) clearSelectionEvent();
        ArrayList elts = document.getElementsInRect(r);
        for (int i = 0; i < elts.size(); i++)
            addElementToSelectionEvent((CommonElement)elts.get(i));
        repaint();
        menuRefreshInvoker.invoke(this);
        
        return selection;
    }
    
    public void selectAllEvent() {
        for (int i = 0; i < document.getElementsCount(); i++)
            addElementToSelectionEvent(document.getElement(i));
        repaint();
        menuRefreshInvoker.invoke(this);
    }
    
    /**
     * Removes element from selection
     */
    public Selection removeElementFromSelectionEvent(CommonElement elt) {
        if (!selection.hasElement(elt)) return selection;
        selection.removeElement(elt);
        menuRefreshInvoker.invoke(this);
        refreshDocumentRelations();
        return selection;
    }
    
    /**
     * Copies elements into clipboard
     */
    public void copySelectionEvent() {
        if (selection.isEmpty()) {
            Log.warn(Lang.get("log.emptyselection"));
            return;
        }
        setClipboard((Selection)selection.clone());
        menuRefreshInvoker.invoke(this);
    }
    
    public void pasteClipboardEvent() {
        if (getClipboard() == null) {
            Log.warn(Lang.get("log.emptyclipboard"));
            return;
        }
        performEvent(new PasteSelectionEvent(this, getClipboard()));
        repaint();
        menuRefreshInvoker.invoke(this);
    }
    
    /**
     * Removes elements from document into selection
     */
    public void cutSelectionEvent() {
        if (selection.isEmpty()) {
            Log.warn(Lang.get("log.emptyselection"));
            return;
        }
        copySelectionEvent();
        dropSelectionEvent();
        menuRefreshInvoker.invoke(this);
    }
    
    /**
     * Removes selection from document
     */
    public void dropSelectionEvent() {
        if (selection.isEmpty()) {
            Log.warn(Lang.get("log.emptyselection"));
            return;
        }
        
        performEvent(new RemoveSelectionEvent(this, selection));
        
        clearSelectionEvent();
        repaint();
        menuRefreshInvoker.invoke(this);
    }    
    
    public void undo() {
        undoRedoStack.undo();
        repaint();
        menuRefreshInvoker.invoke(this);
        refreshDocumentRelations();
    }
    
    public void redo() {
        undoRedoStack.redo();
        repaint();
        menuRefreshInvoker.invoke(this);
        refreshDocumentRelations();
    }
    

    public void groupSelection() {
        if (selection.getElementsCount() < 2)
            Log.warn(Lang.get("log.selectiontoosmall"));
        else {
	        performEvent(new GroupSelectionEvent(this, selection));
	        repaint();
        }
        menuRefreshInvoker.invoke(this);
    }

    public void splitSelection() {
        if (!selection.hasSplitableElements())
            Log.warn(Lang.get("log.notsplitableselection"));
        else {
	        performEvent(new SplitSelectionEvent(this, selection));
	        repaint();
        }
        menuRefreshInvoker.invoke(this);
    } 
    
    public void setDefaultContext() {
        setActiveContext(new DefaultContext(this, parent.getDevice()));
        repaint();
        menuRefreshInvoker.invoke(this);
    }
    
    public void setPaddingContext() {
        setActiveContext(new PaddingContext(this, parent.getDevice()));
        repaint();
        menuRefreshInvoker.invoke(this);
    }

    /**
     * Saves file into specified path
     * @param absoluteFile
     * @throws IOException
     */
    public void saveToFile(File absoluteFile) throws IOException {
        try {
            Serializer.serialize(document, absoluteFile);
            documentPath = absoluteFile.getAbsolutePath();
        } catch (IOException e) {
            throw e;
        }
        documentChanged = false;
    }
    
    public void save() throws IOException {
        try {
            Serializer.serialize(document, new File(documentPath));
        } catch (IOException e) {
            throw e;
        }
        documentChanged = false;
    }
    
    public void openFile(File file) 
    	throws 	JDOMException, IOException, DeserializationException,
    			InstantiationException, IllegalAccessException, Exception {
        
        Document doc = DefaultDeserializer.deserializeDocument(file);
        ArrayList notFoundElements;
        boolean userRefuse = false;
        do {
            notFoundElements = doc.instatiateLibraryElements(librarian, true);
            if (notFoundElements.size() > 0)
                userRefuse = ((Boolean)notFoundElementsInvoker.invoke(notFoundElements)).booleanValue();
        } while ((notFoundElements.size() > 0) && (!userRefuse));
        
        if ((userRefuse) || (doc.instatiateLibraryElements(librarian, true).size() > 0))
            throw new Exception(Lang.get("msg.unabletoopen", file.getAbsolutePath()));
        
        if (!userRefuse) {
            setDocument(doc);
            documentChanged = false;
            documentPath = file.getAbsolutePath();
            clearSelectionEvent();
        }
    }
    
    public Library createEmptyLibrary() {
        Library lib = new Library(Lang.get("application.defaultlibraryname"), "");
        lib.setLibrarian(getLibrarian());
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(lib.getProperties());
        if (changes != null) changes.applyChanges();
        return lib;
    }
    
    public void invokeLibraryPropertiesDialog(Library lib) {
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(lib.getProperties());
        changes.applyChanges();
    }
    
    public void saveLibrary(Library lib, File absoluteFile) throws IOException {
        lib.getProperties().getProperty(XmlConstants.PROP_URL).setValue(absoluteFile.getAbsolutePath());
        Serializer.serialize(lib, absoluteFile);
        saveLibConfig();
    }
    
    public void saveIntoLibrary(String libName, String eltName) {
        getLibrarian().getLibrary(libName).putElement(eltName, getDocument());
        getLibrarian().saveLibraries();
    }
    
    public Librarian getLibrarian() {
        return librarian;
    }
    
    /**
     * @return document properties
     */
    public Properties getDocumentProperties() {
        return getDocument().getProperties();
    }
    
    public void setPropertyDialogInvoker(UIInvoker propertyDialogInvoker) {
        this.propertyDialogInvoker = propertyDialogInvoker;
    }
    
    /**
     * Creates new element from library 
     * @param library
     * @param name
     * @throws DependencyException 
     */
    public void setupCreatingLibraryObject(String library, String name) throws DependencyException {
        CommonElement elt = librarian.getLibrary(library).getNewElement(name);
        LibraryElement libElt = new LibraryElement(elt);
        libElt.getProperties().setProperty(new StringProperty(XmlConstants.PROP_NAME, name));
        libElt.getProperties().setProperty(new StringProperty(XmlConstants.PROP_LIBRARY, library));
        
        setActiveContext(new ElementCreationContext(this, parent.getDevice(), libElt));
    }
    
    public Abilities getAbilities() {
        return new Abilities(document, selection, clipboard, undoRedoStack, librarian);
    }
    
    /**
     * @return list of libraries available in project
     */
    public ArrayList getLibraries() {
        ArrayList result = new ArrayList();
        for (int i = 0; i < getLibrarian().getLibCount(); i++)
            result.add(getLibrarian().getLibrary(i).getProperties().getStringValue(XmlConstants.PROP_NAME));
        return result;
    }
    
    /**
     * @return list of names in specified library
     */
    public ArrayList getElementsInLibrary(String name) {
        return getLibrarian().getLibrary(name).getElementsNames();
    }
    
    public Properties getLibraryProperties(String name) {
        return getLibrarian().getLibrary(name).getProperties();
    }
    
    public Properties getElementInLibraryProperties(String libName, String eltName) {
        return getLibrarian().getLibrary(libName).getElement(eltName).getProperties();
    }
    
    public void removeLibrary(String name) {
        getLibrarian().removeLibrary(name);
        try {
            saveLibConfig();
        } catch (IOException e) {
            Log.warn(Lang.get("log.unabletosavelibconfig"));
        }
    }
    
    public Library loadLibrary(File file) throws JDOMException, IOException, DeserializationException, InstantiationException, IllegalAccessException {
        
        Library lib = DefaultDeserializer.deserializeLibrary(file);
        lib.setLibrarian(getLibrarian());
        return lib;
    }
    
    public void putLibrary(Library lib) {
        getLibrarian().putLibrary(lib);
        try {
            Serializer.serialize(lib, new File(lib.getProperties().getStringValue(XmlConstants.PROP_URL)));
            saveLibConfig();
        } catch (IOException e) {
            Log.warn(Lang.get("log.unabletosavelibconfig"));
        }
    }
    
    public void removeElementInLibrary(String libName, String eltName) {
        Library lib = getLibrarian().getLibrary(libName); 
        lib.removeElement(lib.getElement(eltName).getProperties().getStringValue(XmlConstants.PROP_NAME));
        librarian.saveLibraries();
    }
    
    public void setMenuRefreshInvoker(UIInvoker menuRefreshInvoker) {
        this.menuRefreshInvoker = menuRefreshInvoker;
    }
    
    public void setContextMenuInvoker(UIInvoker contextMenuInvoker) {
        this.contextMenuInvoker = contextMenuInvoker;
    }
    
    public void saveLibConfig() throws IOException {
        Serializer.serialize(getLibrarian(), new File(LIB_CONFIG_PATH));
    }

    public void setLibrarian(Librarian librarian) {
        this.librarian = librarian;
    }
    
    
    public void saveLibraries() {
        librarian.saveLibraries();
    }
    
    public boolean containsLibraryElement(String lib, String elt) {
        return (getLibrarian().getLibrary(lib).getElement(elt) != null);
    }
    
    public void invokeContextMenu(java.awt.Point position) {
        contextMenuInvoker.invoke(position);
    }
    
    public void invokeMenuRefresh() {
        if (menuRefreshInvoker != null) menuRefreshInvoker.invoke(this);
    }
    
    public void sendSelectionFront() {
        performEvent(new SendFrontEvent(this, selection));
        repaint();
    }
    
    public void sendSelectionBack() {
        performEvent(new SendBackEvent(this, selection));
        repaint();
    }
    
    public void export(File file, int format) throws Exception {
        Export.export(file, document, format);
    }
    
    public void invokeViewChange() {
        viewChangeInvoker.invoke(this);
    }

    public UIInvoker getViewChangeInvoker() {
        return viewChangeInvoker;
    }

    public void setViewChangeInvoker(UIInvoker viewChangeInvoker) {
        this.viewChangeInvoker = viewChangeInvoker;
    }
    
    public void setZoomInContext() {
        setActiveContext(new ZoomContext(this, parent.getDevice()));
    }

    public Point getMousePosition() {
        return activeContext.getMousePosition();
    }

    public UIInvoker getMouseChangeInvoker() {
        return mouseChangeInvoker;
    }

    public void setMouseChangeInvoker(UIInvoker mouseChangeInvoker) {
        this.mouseChangeInvoker = mouseChangeInvoker;
    }   
    
    public void invokeMouseChange() {
        mouseChangeInvoker.invoke(this);
    }
    
    public void alignSelection(int direction) {
        if (selection.getElementsCount() < 2)
            Log.warn(Lang.get("log.selectiontoosmall"));
        else {
            performEvent(new AlignEvent(this, selection, direction));
            repaint();
        }
        menuRefreshInvoker.invoke(this);
    }

    public Properties getDefaultProperties() {
        return defaultProperties;
    }
    
    public UIInvoker getDefaultPropertyChangeInvoker() {
        return defaultPropertyChangeInvoker;
    }

    public void setDefaultPropertyChangeInvoker(
            UIInvoker defaultPropertyChangeInvoker) {
        this.defaultPropertyChangeInvoker = defaultPropertyChangeInvoker;
    }
    
    public void invokeDefaultPropertyChange() {
        defaultPropertyChangeInvoker.invoke(defaultProperties);
    }
    
    public void setSelectionProperty(String propertyName, Object newValue) {
        
        PropertyChangeList changeList = new PropertyChangeList();
        for (int i = 0; i < selection.getElementsCount(); i++) {
            CommonElement elt = selection.getElement(i);
            Property p = elt.getProperties().getProperty(propertyName);
            if (p != null) changeList.add(new PropertyChangeLog(p, newValue)); 
        }
        
        if (changeList.size() > 0)
            performEvent(new PropertiesChangeEvent(this, changeList));
        
    }
    
    public void setDefaulProperty(String propertyName, Object newValue) {
        getDefaultProperties().getProperty(propertyName).setValue(newValue);
        setSelectionProperty(propertyName, newValue);
        repaint();
        invokeDefaultPropertyChange();
    }
    
    public void invokeFontStyleDialog() {
        PropertyDescriptor propertyDescriptor = new PropertyDescriptor();
        
        SelectPropertyDescription fontDesc = new SelectPropertyDescription(XmlConstants.PROP_FONT_FACE, Lang.get("property.font.face"), Lang.get("category.style"), 21);
        DefaultArrays.fillFontNames(fontDesc);

        propertyDescriptor.setDescription(fontDesc);
        propertyDescriptor.setDescription(XmlConstants.PROP_FONT_SIZE, Lang.get("property.font.size"), Lang.get("category.style"), 22);

        SelectPropertyDescription alignDesc = new SelectPropertyDescription(XmlConstants.PROP_ALIGN, Lang.get("property.align"), Lang.get("category.style"), 23);
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_LEFT), Lang.get("property.align.left"));
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_CENTER), Lang.get("property.align.center"));
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_RIGHT), Lang.get("property.align.right"));
        propertyDescriptor.setDescription(alignDesc);
        
        propertyDescriptor.setDescription(XmlConstants.PROP_BOLD, Lang.get("property.font.bold"), Lang.get("category.style"), 25);
        propertyDescriptor.setDescription(XmlConstants.PROP_ITALIC, Lang.get("property.font.italic"), Lang.get("category.style"), 26);
        propertyDescriptor.setDescription(XmlConstants.PROP_UNDERLINE, Lang.get("property.font.underline"), Lang.get("category.style"), 27);
        getDefaultProperties().setPropertyDescriptor(propertyDescriptor);
        
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(defaultProperties);
        if (changes != null) changes.applyChanges();
        
        saveChangesOnSelection(changes);
        invokeDefaultPropertyChange();
        repaint();
    }
    
    public void invokeLineStyleDialog() {
        PropertyDescriptor propertyDescriptor = new PropertyDescriptor();
        
        propertyDescriptor.setDescription(XmlConstants.PROP_COLOR, Lang.get("property.color"), Lang.get("category.style"), 20);
        propertyDescriptor.setDescription(XmlConstants.PROP_BACKGROUND, Lang.get("property.background"), Lang.get("category.style"), 21);
        propertyDescriptor.setDescription(XmlConstants.PROP_LINE_WIDTH, Lang.get("property.linewidth"), Lang.get("category.style"), 22);
        
        SelectPropertyDescription lineStyleDesc = new SelectPropertyDescription(XmlConstants.PROP_LINE_STYLE, Lang.get("property.linestyle"), Lang.get("category.style"), 23);
        DefaultArrays.fillLineStyles(lineStyleDesc);
        propertyDescriptor.setDescription(lineStyleDesc);

        getDefaultProperties().setPropertyDescriptor(propertyDescriptor);
        
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(defaultProperties);
        if (changes != null) changes.applyChanges();
        
        saveChangesOnSelection(changes);
        invokeDefaultPropertyChange();
        repaint();
    }
    
    public void invokeSelectionPropertiesDialog() {
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(selection.getProperties());
        
        saveChangesOnSelection(changes);
        invokeDefaultPropertyChange();
        repaint();
    }
    
    private void saveChangesOnSelection(PropertyChangeList changes) {
        PropertyChangeList selectionChanges = new PropertyChangeList();
        
        if (changes == null) return;
        
        for (int i = 0; i < changes.size(); i++) {
            for (int j = 0; j < selection.getElementsCount(); j++) {
                CommonElement elt = selection.getElement(j);
                Property p = elt.getProperties().getProperty(changes.get(i).getProperty().getName());
                if (p != null) {
                    Object value = changes.get(i).getNewValue();
                    if (value instanceof Cloneable) {
                        try {
                            // creates clone of an object
                            value = value.getClass().getMethod("clone", null).invoke(value, null);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                    selectionChanges.add(new PropertyChangeLog(p, value));
                }
                    
            }
        }
        
        if (selectionChanges.size() > 0)
            performEvent(new PropertiesChangeEvent(this, selectionChanges));
        selectionChanges.applyChanges();
        
    }
    
    public void invokeDocumentPropertiesDialog() {
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(document.getProperties());
        if (changes != null) changes.applyChanges();
        repaint();
        invokeViewChange();
    }
    
    public void invokeGridPropertiesDialog() {
        PropertyDescriptor oldPropDesc = getDocument().getProperties().getPropertyDescriptor();
        getDocument().getProperties().setPropertyDescriptor(gridPropertyDescriptor);
        PropertyChangeList changes = (PropertyChangeList)propertyDialogInvoker.invoke(document.getProperties());
        getDocument().getProperties().setPropertyDescriptor(oldPropDesc);
        if (changes != null) changes.applyChanges();
        repaint();
        invokeViewChange();
    }
    
    public boolean hasSelectionElement(CommonElement elt) {
        return selection.hasElement(elt);
    }
    
    public Selection getSelection() {
        return selection;
    }

    public UIInvoker getNotFoundElementsInvoker() {
        return notFoundElementsInvoker;
    }

    public void setNotFoundElementsInvoker(UIInvoker notFoundElementsInvoker) {
        this.notFoundElementsInvoker = notFoundElementsInvoker;
    }
    
    public void addGuideLine(GuideLine guideLine) {
        getDocument().addGuideLine(guideLine);
    }
    
    public void removeGuideLine(GuideLine guideLine) {
        getDocument().removeGuideLine(guideLine);
    }
    
    public boolean hasGuideLine(GuideLine guideLine) {
        return getDocument().hasGuideLine(guideLine);
    }
    
    public String getContextInfo() {
        return activeContext.getContextInfo();
    }
    
    public RelationManager getRelationManager() {
        return relationManager;
    }
    
    public boolean isDocumentChanged() {
        return documentChanged;
    }

    public String getDocumentPath() {
        return documentPath;
    }
    
    public void refreshDocumentRelations() {
        selection.getRelations().clear();
        if (!isSnapTech()) return;
        
        relationManager.updateRelations();
        for (int i = 0 ; i < getDocument().getElementsCount(); i++) {
            CommonElement elt = getDocument().getElement(i);
            if (selection.hasElement(elt)) continue;
            if (getDocument().getElement(i) instanceof SubmissiveElement) {
                ((SubmissiveElement)getDocument().getElement(i)).askDominant(selection);
            }
        }
        
    }
    
    public boolean emptyDocument() {
        return (getDocument().getElementsCount() == 0);
    }
}