/*
 * Created on Apr 7, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.contexts;

import java.awt.event.MouseEvent;

import diagnose.graphicengine.Manager;
import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.ComposedElement;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.graphicengine.elements.system.AbstractControlPointElement;
import diagnose.graphicengine.elements.system.GuideLine;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.CursorFactory;
import diagnose.utils.DPIUtils;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.serialization.XmlConstants;

/**
 * This context is default and reacts on base events as keypressing
 * and selection creation
 */
public class DefaultContext extends SelectingContext {
    
    private static final int DECIMAL_CORRECTION = 100;
    
    /**
     * In this fields is currently manipulated object
     */
    private VisualElement activeElement;
    
    /**
     * Point where was active element selected 
     */
    private Point selectedAt;

    /**
     * Contains point where was active element originaly selected
     */
    private Point selectedOriginalyAt;
    
    private boolean ignoreClickEvent = false;
    
    public DefaultContext(Manager manager, AbstractDevice device) {
        super(manager, device);
    }
    
    public void activateContext() {
        getManager().setCursor(CursorFactory.getDefaultCursor());
    }
    
    public void mouseClicked(MouseEvent e) {
        
        super.mouseClicked(e);
        
        if (ignoreClickEvent) {
            ignoreClickEvent = false;
            return;
        }
        
        Point p = viewCorrection(DPIUtils.pixelsToMilimeters(e.getPoint()));
        
        VisualElement elt = getManager().getDocument().receivesMouseEvent(p, getDevice());        
        // runs property dialog
        if ((elt != null) && (elt instanceof CommonElement) && (e.getClickCount() == 2)) {
            getManager().invokeSelectionPropertiesDialog();
            getManager().clearSelectionEvent();
            return;
        }
            
            
        if ((elt == null) && ((!e.isShiftDown() && !e.isControlDown()))) {
            getManager().clearSelectionEvent();
        } else if (elt instanceof CommonElement) {
            
            CommonElement celt = (CommonElement)elt;
            
            if (!(e.isShiftDown() || e.isControlDown())) {
                getManager().clearSelectionEvent();
                getManager().addElementToSelectionEvent(celt);
            }
            
            if (e.isShiftDown()) getManager().addElementToSelectionEvent(celt);
            
            if (e.isControlDown()) {
                if (!getManager().getSelection().hasElement(celt))
                    getManager().addElementToSelectionEvent(celt);
                else
                    getManager().removeElementFromSelectionEvent(celt);
            }   
        }
        
        getManager().repaint();
    }
    
    public void mouseDragged(MouseEvent e) {
        
        super.mouseDragged(e);
        
        Point ml = viewCorrection(DPIUtils.pixelsToMilimeters(e.getPoint()));
        
        if (selectedAt == null) return;
        
        double xMove = ml.x - selectedAt.x;
        double yMove = ml.y - selectedAt.y;
    
    	if (activeElement != null) {
    	    /////
            if (activeElement instanceof CommonElement)
                if (getManager().getSelection().hasElement((CommonElement)activeElement))
                    activeElement = getManager().getSelection();
            /////
    	    if ((activeElement != getManager().getSelection()) 
                    && (activeElement instanceof CommonElement)
                    && (!(activeElement instanceof GuideLine))) {
    	        getManager().clearSelectionEvent();
    	        getManager().addElementToSelectionEvent((CommonElement)activeElement);
    	    }
    	    
	        activeElement.move(new Point(xMove, yMove));

            if (activeElement instanceof ComposedElement) {
                if (getManager().isSnapTech())
                    getManager().getRelationManager().updateRelations((ComposedElement)activeElement);
            } else if (activeElement instanceof SubmissiveElement) {
                if (getManager().isSnapTech())
                    getManager().getRelationManager().updateRelations((SubmissiveElement)activeElement);
            }
            else if (activeElement instanceof AbstractControlPointElement) {
                if (getManager().isSnapTech()) {
                    CommonElement controled = ((AbstractControlPointElement)activeElement).getControls();
                    if (controled instanceof SubmissiveElement)
                        getManager().getRelationManager().updateRelations((SubmissiveElement)controled);
                }
            }
            
	        selectedAt.translate(xMove, yMove);
	        
	        Point snapCorrection = getSnappingCorrection(activeElement);
	        
	        if ((snapCorrection != null) && 
	                (snapCorrection.distance(0, 0) < getManager().getSnapPrecissness())) {
	            activeElement.move(snapCorrection);
	            selectedAt.translate(snapCorrection.x, snapCorrection.y);
	        } 
            
            if (activeElement instanceof GuideLine) {
                GuideLine guideLine = (GuideLine)activeElement;
                int orientation = guideLine.getProperties().getIntValue(XmlConstants.PROP_ORIENTATION);
                Point pos = guideLine.getControlPoint(0);
                if 
                    (((orientation == GuideLine.HORIZONTAL) && (pos.y < -getManager().getPadding().y / getManager().getZoom()))
                 || ((orientation == GuideLine.VERTICAL) && (pos.x < -getManager().getPadding().x / getManager().getZoom()))) {
                    getManager().getDocument().removeGuideLine(guideLine);
                    activeElement = null;   
                    selectedOriginalyAt = (Point)selectedAt.clone();
                }
            }
    	} else {
    	    selectionArea = new Rect(selectedOriginalyAt, ml);
    	}
        
        getManager().repaint();
        
    }
   
    public void mousePressed(MouseEvent e) {
        
        super.mousePressed(e);

        if (e.isPopupTrigger()) {
            getManager().invokeContextMenu(e.getPoint());
        }
        
        Point p = viewCorrection(DPIUtils.pixelsToMilimeters(e.getPoint()));
        
        activeElement = getManager().getSelection().getControlPoints().receivesMouseEvent(p, getDevice());
        if (activeElement == null) 
            activeElement = (getManager().getSelection().receivesMouseEvent(p, getDevice()) == null ? null : getManager().getSelection());

        if (activeElement == null)
            activeElement = getManager().getDocument().receivesMouseEvent(p, getDevice());
        
        selectedAt = p;
        selectedOriginalyAt = (Point)p.clone();        
    }
    
    public void mouseReleased(MouseEvent e) {
        
        super.mouseReleased(e);
        
        if (e.isPopupTrigger()) {
            getManager().invokeContextMenu(e.getPoint());
        }
        
        if (activeElement != null) {
	        Point ml = selectedAt;
	        double xMove = ml.x - selectedOriginalyAt.x;
	        double yMove = ml.y - selectedOriginalyAt.y;
	
	        activeElement.move(new Point(-xMove, -yMove));
	        getManager().moveElement(activeElement, new Point(xMove, yMove));
	        activeElement = null;
        }
        
        selectedAt = null;
        selectedOriginalyAt = null;
        
        
        if (selectionArea != null) {
            if ((e.getModifiers() & MouseEvent.SHIFT_MASK) > 0) {
                getManager().setSelectionByRect(selectionArea, true);
            } else { 
                getManager().setSelectionByRect(selectionArea, false); 
            }
            ignoreClickEvent = true;
        }
        selectionArea = null;
        
        getManager().repaint();
    }
    
    
    public String getContextInfo() {
        if (selectionArea != null) 
            return Lang.get("context.selectedarea",
                            round(selectionArea.getWidth()),
                            round(selectionArea.getHeight()));
        else return Lang.get("context.selection");
    }
    
    private static String round(double value) {
        return Double.toString(Math.rint((value * DECIMAL_CORRECTION)) / DECIMAL_CORRECTION);
    }
}
