/*
 * Created on Apr 7, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.contexts;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.ArrayList;

import diagnose.graphicengine.Manager;
import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.ComposedElement;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.Document;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.utils.DPIUtils;
import diagnose.utils.geometry.Point;

/**
 * Parent class for transformation events from GUI to logical events
 */
public abstract class EventContext {

    private Manager manager;
    private Point mousePosition = null;
    private AbstractDevice device;
    
    public EventContext(Manager manager, AbstractDevice device) {
        super();
        this.manager = manager;
        this.device = device;
    }
  
    public Manager getManager() {
        return manager;
    }
    
    public void setManager(Manager manager) {
        this.manager = manager;
    }
    
    public AbstractDevice getDevice() {
        return device;
    }
    
    public void setDevice(AbstractDevice device) {
        this.device = device;
    }
    
    /**
     * This function is called after switching context
     */
    public void activateContext() {}
    
    public void keyEvent(int keyCode, int modifiers) { 
        if (keyCode == KeyEvent.VK_ESCAPE) cancelContext();
    }
    
    public void mouseClicked(MouseEvent e) {
        setMousePosition(e.getPoint());
    }
    
    public void mousePressed(MouseEvent e)  { 
        setMousePosition(e.getPoint());
    }
    
    public void mouseReleased(MouseEvent e) {
        setMousePosition(e.getPoint());
    }
    
    public void mouseDragged(MouseEvent e) { 
        setMousePosition(e.getPoint());
    } 
    
    public void mouseMoved(MouseEvent e) { 
        setMousePosition(e.getPoint());
    }
    
    public void mouseEntered(MouseEvent e) {
        setMousePosition(e.getPoint());
    }
    
    public void mouseExited(MouseEvent e) {
        setMousePosition(null);
    }
    
    protected void setMousePosition(java.awt.Point p) {
        if (p != null) {
            Point px = DPIUtils.pixelsToMilimeters(p);
            px = viewCorrection(px);
            mousePosition = px;
        } else mousePosition = null;
        manager.invokeMouseChange();
    }
    
    public void paint(AbstractDevice device) { }
    
    /**
     * @return position of mouse otherwise null
     */
    public Point getMousePosition() {
        return mousePosition;
    }
    
    /**
     * @return informations about context
     */
    public abstract String getContextInfo();
    
    /**
     * This method is called for cleaning up after changing contexts
     */
    public void cancelContext() { }
    
    /**
     * @return direction of snapping correction
     * @param snapping represents VisualElement or Point that should be snapped
     */
    protected Point getSnappingCorrection(Object snapping) {

        VisualElement activeElement = null;
        ArrayList snapPoints = new ArrayList();
        
        if (snapping != null) {
	        if (snapping instanceof VisualElement) {
	            activeElement = (VisualElement)snapping;
	            snapPoints = activeElement.getSnapablePoints();
	        } else {
	            snapPoints = new ArrayList();
	            snapPoints.add(snapping);
	        }
        }
        
        Document doc = getManager().getDocument();
        Point bestPoint = null; // snapping correction
        Point np = null;
        double newDist = 0;
        double bestDist = Integer.MAX_VALUE;
        
        for (int i = 0; i < snapPoints.size(); i++) {
            Point p = (Point)snapPoints.get(i);
            
            // snapping to grid
            if (getManager().isSnapToGrid())
            {
                double gridSize = getManager().getGridSize();
                np = new Point(
                    Math.round(p.x / gridSize) * gridSize,
                    Math.round(p.y / gridSize) * gridSize);
                
                newDist = p.distance(np);
                if ((newDist < bestDist) && (newDist > 0) && (newDist < manager.getSnapPrecissness())) {
	                bestDist = newDist;
	                bestPoint = np;
	                bestPoint.x -= p.x;
	                bestPoint.y -= p.y;
                }
            }
            
            // snapping to guide lines
            if (getManager().isSnapToGuideLines()) {
                np = getManager().getDocument().getGuideLines().snaps(p);
                if (np != null) newDist = p.distance(np);
                else newDist = Double.MAX_VALUE;
                if ((newDist < bestDist) && (newDist > 0) && (newDist < manager.getSnapPrecissness())) {
                    bestPoint = np;
                    bestDist = newDist;
                    bestPoint.x -= p.x;
                    bestPoint.y -= p.y;
                }
            }
            
            // snapping to other elements
            if (getManager().isSnapToElements()) {
                for (int j = 0; j < doc.getElementsCount(); j++) {
                    
                    VisualElement elt = doc.getElement(j);
                    
                    if (elt == activeElement) continue;
                    if ((activeElement instanceof ComposedElement) &&
                            (elt instanceof CommonElement) &&
                            ((ComposedElement)activeElement).containsElement((CommonElement)elt)) continue;
                
    	            np = elt.snaps(p); 
    	           
    	            if (np != null) {
    	                newDist = p.distance(np);
    	                if ((newDist < bestDist) && (newDist > 0) && (newDist < manager.getSnapPrecissness())) {
    		                bestDist = newDist;
    		                bestPoint = np;
    		                bestPoint.x -= p.x;
    		                bestPoint.y -= p.y;
    	                }
    	            }
                }
            }
        }
        return bestPoint;
    }
    
    /**
     * Gets point from mouseevent an returns new Point with zoomCorrection
     * and padding correction
     */
    protected Point viewCorrection(MouseEvent e) {
        Point p = new Point(e.getPoint());
        return viewCorrection(p);
    }
    
    /**
     * @return new Point with zoomCorrection and padding correction
     */
    protected Point viewCorrection(Point p) {
        Point padding = getManager().getPadding();
        p = (Point)p.clone();
        p.translate(-padding.x, -padding.y);
        p.multiply(1 / getManager().getZoom());
        return p;
    }
}
