/*
 * Created on Apr 7, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.devices;

import java.awt.Color;
import java.awt.Font;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

import diagnose.utils.DPIUtils;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;

/**
 * Interface between GUI and internal logic
 */
public abstract class AbstractDevice {

    public static final int IGNORE_VIEW_SETTINGS = 0x00;
    public static final int USE_ZOOM    = 0x01;
    public static final int USE_PADDING = 0x02;
    public static final int DEFAULT_DRAW = 0x03;
    
    public static final int CLEAR_LINE      = 1;
    public static final int POINT_LINE      = 2;
    public static final int DASH_LINE       = 3;
    public static final int DASH_POINT_LINE = 4;
    
    public static final int FONT_NORMAL     = 0x00;
    public static final int FONT_BOLD       = 0x01;
    public static final int FONT_ITALIC     = 0x02;
    public static final int FONT_UNDERLINE  = 0x04;
    
    public static final int ALIGN_LEFT      = 1;
    public static final int ALIGN_RIGHT     = 2;
    public static final int ALIGN_CENTER    = 3;
    
    public static double DEFAULT_LINE_WIDTH = 0.35; // mm
    public static int DEFAULT_FONT_SIZE = 12;
    public static String DEFAULT_FONT_FACE = "SansSerif";
    
    private Color fgColor = Color.BLACK;
    private double zoom = 1;
    private Point padding = new Point(0, 0);
    private double lineWidth = 0.35; // mm
    private int lineStyle = CLEAR_LINE;
    private int fontStyle = FONT_NORMAL;
    private int textAlign = ALIGN_LEFT;
    private int fontSize = 12;
    private String fontFamily = Lang.get("application.defaultfont");
    
    public AbstractDevice() {
        setDefaultValues();
    }
    
    public void setDefaultValues() {
        setColor(Color.BLACK);
        setFontFamily(DEFAULT_FONT_FACE);
        setFontSize(DEFAULT_FONT_SIZE);
        setTextAlign(ALIGN_LEFT);
        setLineStyle(CLEAR_LINE);
        setLineWidth(DEFAULT_LINE_WIDTH);
    }
    
    public Color getColor() {
        return fgColor;
    }
    
    public void setColor(Color fgColor) {
        this.fgColor = fgColor;
    }
    
    public Point getPadding() {
        return padding;
    }
    
    public void setPadding(Point padding) {
        this.padding = padding;
    }
    
    public double getZoom() {
        return zoom;
    }
    
    public void setZoom(double zoom) {
        this.zoom = zoom;
    }
    
    public int getLineStyle() {
        return lineStyle;
    }

    public void setLineStyle(int lineStyle) {
        this.lineStyle = lineStyle;
    }
    
    public int getFontStyle() {
        return fontStyle;
    }

    public void setFontStyle(int fontStyle) {
        this.fontStyle = fontStyle;
    }

    public int getTextAlign() {
        return textAlign;
    }

    public void setTextAlign(int textAlign) {
        this.textAlign = textAlign;
    }

    /**
     * Sets the line width in milimeters
     */
    public void setLineWidth(double width) {
        this.lineWidth = width;
    }
    
    /**
     * Returns line width in milimeters
     */
    public double getLineWidth() {
        return lineWidth;
    }
    
    public int getFontSize() {
        return fontSize;
    }

    public void setFontSize(int fontSize) {
        this.fontSize = fontSize;
    }

    public String getFontFamily() {
        return fontFamily;
    }

    public void setFontFamily(String fontFamily) {
        this.fontFamily = fontFamily;
    }

    /**
     * @return width of visible part of device (in milimeters), in case device has
     * unlimited width returns -1 
     */
    public abstract double getVisibleWidth();
    
    /**
     * @return height of visible part of device (in milimeters), in case device has
     * unlimited height returns -1 
     */
    public abstract double getVisibleHeight();

    public abstract void line(double x1, double y1, double x2, double y2, int modifiers);
    public abstract void rect(double x, double y, double width, double height, int modifiers);
    public abstract void fillRect(double x, double y, double width, double height, int modifiers);
    public abstract void oval(double x, double y, double width, double height, int modifiers);
    public abstract void fillOval(double x, double y, double width, double height, int modifiers);
    public abstract void polygon(double [] x, double [] y, int modifiers);
    public abstract void fillPolygon(double [] x, double [] y, int modifiers);
    
    public void line(Point p1, Point p2, int modifiers) {
        line(p1.x, p1.y, p2.x, p2.y, modifiers);
    }
    
    public void rect(Rect r, int modifiers) {
        Point p = r.getPosition();
        rect(p.x, p.y, r.getWidth(), r.getHeight(), modifiers);
    }
    
    public void fillRect(Rect r, int modifiers) {
        Point p = r.getPosition();
        fillRect(p.x, p.y, r.getWidth(), r.getHeight(), modifiers);
    }
    
    public void fillOval(Rect r, int modifiers) {
        Point p = r.getPosition();
        fillOval(p.x, p.y, r.getWidth(), r.getHeight(), modifiers);
    }
    
    public void oval(Rect r, int modifiers) {
        Point p = r.getPosition();
        oval(p.x, p.y, r.getWidth(), r.getHeight(), modifiers);
    }

    public void text(String text, Rect r, int modifiers) {
        Point p = r.getPosition();
        text(text, p.x, p.y, r.getWidth(), r.getHeight(), modifiers);
    }
    
    protected boolean isUseZoom(int modifiers) {
        return (modifiers & USE_ZOOM) > 0;
    }
    
    protected boolean isUsePadding(int modifiers) {
        return (modifiers & USE_PADDING) > 0;
    }
    
    /**
     * Outputs one line of text into specified position (in pixels)
     * using fontsize
     */
    protected abstract void outputTextLine(double x, double y, String line, int modifiers);
    
    /** 
     * @return rectangle covered by text
     */
    protected abstract Rectangle2D getTextBounds(String line);
    
    public void text(String text, double x, double y, double width, double height, int modifiers) {
                 
        x = DPIUtils.milimetersToPixels(x);
        y = DPIUtils.milimetersToPixels(y);
        width = DPIUtils.milimetersToPixels(width);
        height = DPIUtils.milimetersToPixels(height);
                
        double fontHeight = getTextBounds(text).getHeight();
        
        ArrayList lines = new ArrayList();
        String [] baseLines = text.split("\n");
        for (int i = 0; i < baseLines.length; i++) {
            String [] words = baseLines[i].split(" ");
            
            String line = words[0];
            for (int j = 1; j < words.length; j++) {
                if (getTextBounds(line + " " + words[j]).getWidth() > width) {
                    lines.add(line);
                    line = words[j];
                } else {
                    line += " " + words[j];
                }
            }
            if ((line.length() > 0) || (words.length == 1)) lines.add(line);
        }
        
        for (int i = 0; i < lines.size(); i++) {
            double actLine = i * fontHeight * 1.2; // line spacing
            if (actLine + fontHeight > height) return;
            
            String line = (String)lines.get(i);
            
            double xCorrection = 0;
            double textWidth = getTextBounds(line).getWidth();
            if ((getTextAlign() == ALIGN_CENTER) || (getTextAlign() == ALIGN_RIGHT)) { 
                
                if (getTextAlign() == ALIGN_CENTER) xCorrection = (width - textWidth) / 2;
                if (getTextAlign() == ALIGN_RIGHT) xCorrection = (width - textWidth);
            }
            
            outputTextLine(x + xCorrection, y + actLine + fontHeight, line, modifiers);
        }
    }
    
    protected Font getFont() {
        int style = 0;
        if ((getFontStyle() & FONT_BOLD) > 0) style |= Font.BOLD;
        if ((getFontStyle() & FONT_ITALIC) > 0) style |= Font.ITALIC;
        Font font = new Font(getFontFamily(), style, getFontSize());
        return font;
    }
}