/*
 * Created on Jun 17, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.devices;

import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;

import javax.swing.JComponent;

import diagnose.graphicengine.contexts.EventContext;
import diagnose.utils.DPIUtils;
import diagnose.utils.geometry.Point;

/**
 *
 */
public class CanvasDevice extends InteractiveDevice {

    private Graphics2D canvas;
    private JComponent owner;
    
    public CanvasDevice() {
        super(null);
    }

    public CanvasDevice(JComponent owner, EventContext eventContext) {
        super(eventContext);
        this.canvas = (Graphics2D)owner.getGraphics();
        this.owner = owner;
        owner.addMouseListener(new MouseStaticEvents());
        owner.addMouseMotionListener(new MouseMotionEvents());
        owner.addKeyListener(new KeyPressEvents());
    }
    
    public void setCanvas(Graphics2D canvas) {
        this.canvas = canvas;
    }
    
    public void line(double x1, double y1, double x2, double y2, int modifiers) {
        
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        x1 = DPIUtils.milimetersToPixels(x1);
        y1 = DPIUtils.milimetersToPixels(y1);
        x2 = DPIUtils.milimetersToPixels(x2);
        y2 = DPIUtils.milimetersToPixels(y2);
        
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.drawLine((int)x1, (int)y1, (int)x2, (int)y2);
        revertTransforms(transform);
    }
    
    public void rect(double x, double y, double width, double height, int modifiers) {
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        x = DPIUtils.milimetersToPixels(x);
        y = DPIUtils.milimetersToPixels(y);
        width = DPIUtils.milimetersToPixels(width);
        height = DPIUtils.milimetersToPixels(height);

        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.drawRect((int)x, (int)y, (int)width, (int)height);
        revertTransforms(transform);
    }
    
    public void fillRect(double x, double y, double width, double height, int modifiers) {
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        x = DPIUtils.milimetersToPixels(x);
        y = DPIUtils.milimetersToPixels(y);
        width = DPIUtils.milimetersToPixels(width);
        height = DPIUtils.milimetersToPixels(height);
        
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.fillRect((int)x, (int)y, (int)width, (int)height);
        revertTransforms(transform);
    }
    
    public void oval(double x, double y, double width, double height, int modifiers) {
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        x = DPIUtils.milimetersToPixels(x);
        y = DPIUtils.milimetersToPixels(y);
        width = DPIUtils.milimetersToPixels(width);
        height = DPIUtils.milimetersToPixels(height);
        
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.drawOval((int)x, (int)y, (int)width, (int)height);
        revertTransforms(transform);
    }
    
    public void fillOval(double x, double y, double width, double height, int modifiers) {
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        x = DPIUtils.milimetersToPixels(x);
        y = DPIUtils.milimetersToPixels(y);
        width = DPIUtils.milimetersToPixels(width);
        height = DPIUtils.milimetersToPixels(height);
        
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.fillOval((int)x, (int)y, (int)width, (int)height);
        revertTransforms(transform);
    }
    
    public void polygon(double [] x, double [] y, int modifiers) {
        
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        int [] _x = new int[x.length];
        int [] _y = new int[y.length];
        for (int i = 0; i < x.length; i++) {
            _x[i] = (int)DPIUtils.milimetersToPixels(x[i]);
            _y[i] = (int)DPIUtils.milimetersToPixels(y[i]);
        }
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.drawPolygon(_x, _y, _x.length);
        revertTransforms(transform);
    }
    
    public void fillPolygon(double [] x, double [] y, int modifiers) {
        
        canvas.setColor(getColor());
        canvas.setStroke(getStroke());
        
        int [] _x = new int[x.length];
        int [] _y = new int[y.length];
        for (int i = 0; i < x.length; i++) {
            _x[i] = (int)DPIUtils.milimetersToPixels(x[i]);
            _y[i] = (int)DPIUtils.milimetersToPixels(y[i]);
        }
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.fillPolygon(_x, _y, _x.length);
        revertTransforms(transform);
    }
    
    /**
     * Apply padding and zoom transformation
     * @return transformation previous the applications
     */
    private AffineTransform applyBasicTransforms(int modifiers) {
        AffineTransform transform = canvas.getTransform();

        if (isUsePadding(modifiers)) {
            Point padding = DPIUtils.milimetersToPixels(getPadding());
            canvas.translate(padding.x, padding.y);
        }
        
        if (isUseZoom(modifiers)) {
            double zoom = getZoom();
            canvas.scale(zoom, zoom);
        }
        
        return transform;
    }
    
    private void revertTransforms(AffineTransform transform) {
        canvas.setTransform(transform);
    }

    protected class MouseStaticEvents implements MouseListener {

        public void mouseClicked(MouseEvent e) {
            getEventContext().mouseClicked(e);
        }

        public void mouseEntered(MouseEvent e) {
            getEventContext().mouseEntered(e);
        }
        
        public void mouseExited(MouseEvent e) {
            getEventContext().mouseExited(e);
        }

        public void mousePressed(MouseEvent e) {
            getEventContext().mousePressed(e);
        }
        

        public void mouseReleased(MouseEvent e) {
            getEventContext().mouseReleased(e);
        }
    }
    
    protected class MouseMotionEvents implements MouseMotionListener {

        public void mouseDragged(MouseEvent e) {
            getEventContext().mouseDragged(e);
        }

        public void mouseMoved(MouseEvent e) {
            getEventContext().mouseMoved(e);
        }
    }

    protected class KeyPressEvents implements KeyListener {

        public void keyTyped(KeyEvent e) {
        }

        public void keyPressed(KeyEvent e) {
            getEventContext().keyEvent(e.getKeyCode(), e.getModifiers());
        }

        public void keyReleased(KeyEvent e) {
        }  
    }

    public double getVisibleWidth() {
        return DPIUtils.pixelsToMilimeters(owner.getWidth());
    }

    public double getVisibleHeight() {
        return DPIUtils.pixelsToMilimeters(owner.getHeight());
    }
    
    /**
     * @return stroke with attributes specified for device
     */
    protected Stroke getStroke() {

        float width = (float)DPIUtils.milimetersToPixels(getLineWidth());
        if (getLineStyle() == CLEAR_LINE) return new BasicStroke(width, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
        
        if (getLineStyle() == DASH_LINE) return new BasicStroke(width, BasicStroke.CAP_ROUND,
                BasicStroke.JOIN_ROUND, 0, new float[]{20,10,20,10}, 0);
        
        if (getLineStyle() == DASH_POINT_LINE) return new BasicStroke(width, BasicStroke.CAP_ROUND,
                BasicStroke.JOIN_ROUND, 0, new float[]{20,10,1,10}, 0);
        
        if (getLineStyle() == POINT_LINE) return new BasicStroke(width, BasicStroke.CAP_ROUND,
                BasicStroke.JOIN_ROUND, 0, new float[]{0,6,0,6}, 0);
        
        return new BasicStroke(width);
    }

    protected void outputTextLine(double x, double y, String line, int modifiers) {
        AffineTransform transform = applyBasicTransforms(modifiers);
        canvas.setColor(getColor());  
        canvas.setFont(getFont());
        canvas.drawString(line, (float)x, (float)y);
        
        double textWidth = canvas.getFontMetrics().stringWidth(line);
        
        if ((getFontStyle() & FONT_UNDERLINE) > 0) {
            setLineStyle(CLEAR_LINE);
            setLineWidth(DEFAULT_LINE_WIDTH);
            canvas.setColor(getColor());
            canvas.setStroke(getStroke());
            canvas.drawLine((int)x, (int)(y + 2), 
                    (int)(x + textWidth), (int)(y + 2));
        }
        revertTransforms(transform);
    }

    protected Rectangle2D getTextBounds(String line) {
        return getFont().getStringBounds(line, canvas.getFontRenderContext());
    }
}
