/*
 * Created on Jun 17, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.devices;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.Writer;

import org.jdom.DefaultJDOMFactory;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import diagnose.utils.DPIUtils;

/**
 * Exports document into SVG format
 */
public class SVGExportDevice extends ExportDevice {
    
    private static DefaultJDOMFactory builder = new DefaultJDOMFactory();
    private Document doc;
    private Element rootElement;
    private static final Graphics2D virtualDevice = GraphicsEnvironment.getLocalGraphicsEnvironment().createGraphics(new BufferedImage(100, 100, BufferedImage.TYPE_INT_RGB));


    /**
     * @param writer
     */
    public SVGExportDevice(Writer writer, double width, double height) {
        super(writer);
        rootElement = builder.element("svg");
        rootElement.setAttribute("width", getOutputValue(width));
        rootElement.setAttribute("height", getOutputValue(height));
        doc = builder.document(rootElement);
    }

    public void line(double x1, double y1, double x2, double y2, int modifiers) {

        Element elt = builder.element("line");
        elt.setAttribute("style", getStroke());
        elt.setAttribute("x1", getOutputValue(x1 * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("y1", getOutputValue(y1 * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("x2", getOutputValue(x2 * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("y2", getOutputValue(y2 * (isUseZoom(modifiers) ? getZoom() : 1)));
        rootElement.addContent(elt);
    }
    
    protected void baseRect(double x, double y, double width, double height,
            int modifiers, String style) {
        
        Element elt = builder.element("rect");
        elt.setAttribute("style", style);
        elt.setAttribute("x", getOutputValue(x * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("y", getOutputValue(y * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("width", getOutputValue(width * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("height", getOutputValue(height * (isUseZoom(modifiers) ? getZoom() : 1)));
        rootElement.addContent(elt);
    }

    public void rect(double x, double y, double width, double height,
            int modifiers) {
        baseRect(x, y, width, height, modifiers,
                getStroke() + "fill: none;");
    }

    public void fillRect(double x, double y, double width, double height,
            int modifiers) {
        baseRect(x, y, width, height, modifiers,
                getStroke() + "fill: " + colorToString(getColor()) + ";");
    }
    
    protected void baseOval(double x, double y, double width, double height,
            int modifiers, String style) {
        Element elt = builder.element("ellipse");
        elt.setAttribute("style", style);
        double cx = x + width / 2;
        double cy = y + height / 2;
        double rx = width / 2;
        double ry = height / 2;
        elt.setAttribute("cx", getOutputValue(cx * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("cy", getOutputValue(cy * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("rx", getOutputValue(rx * (isUseZoom(modifiers) ? getZoom() : 1)));
        elt.setAttribute("ry", getOutputValue(ry * (isUseZoom(modifiers) ? getZoom() : 1)));
        rootElement.addContent(elt);
    }

    public void oval(double x, double y, double width, double height,
            int modifiers) {
        baseOval(x, y, width, height, modifiers,
                getStroke() + " fill: none;");
    }
    
    public void fillOval(double x, double y, double width, double height,
            int modifiers) {
        baseOval(x, y, width, height, modifiers,
                getStroke() + " fill: " + colorToString(getColor()) + ";");
    }
    
    protected void basePolygon(double[] x, double[] y, int modifiers, String style) {
        StringBuilder points = new StringBuilder();
        Element elt = builder.element("polygon");
        
        for (int i = 0; i < x.length; i++) {
            points.append(getOutputValue(x[i]));
            points.append(",");
            points.append(getOutputValue(y[i]));
            points.append(" ");
        }
        
        elt.setAttribute("points", points.toString().trim());
        elt.setAttribute("style", style);
        rootElement.addContent(elt);
    }

    public void fillPolygon(double[] x, double[] y, int modifiers) {
        basePolygon(x, y, modifiers,
                getStroke() + " fill: " + colorToString(getColor()) + ";");
    }
    
    public void polygon(double[] x, double[] y, int modifiers) {
        basePolygon(x, y, modifiers, getStroke() + " fill: none;");
    }

    public void close() throws IOException {
        XMLOutputter otp = new XMLOutputter(Format.getPrettyFormat());
        otp.output(doc, getWriter());
    }
    
    private static String colorToString(Color color) {
        return "rgb(" + color.getRed() + ", " + color.getGreen() + ", " + color.getBlue() + ")";
    }
    
    public double getVisibleWidth() {
        return -1;
    }

    public double getVisibleHeight() {
        return -1;
    }
    
    private static String getOutputValue(double value) {
        return Double.toString(DPIUtils.milimetersToPixels(value));
    }
    
    private String getStrokeWidth() {
        return Double.toString(DPIUtils.milimetersToPixels(getLineWidth()));
    }
    
    private String getStroke() {
        String result =  "stroke: " + colorToString(getColor()) + "; stroke-width: " + getStrokeWidth() + "px;";
        if (getLineStyle() == DASH_LINE) result += "stroke-dasharray: 20 10 20 10;";
        if (getLineStyle() == DASH_POINT_LINE) result += "stroke-dasharray: 20 10 1 10;";
        if (getLineStyle() == POINT_LINE) result += "stroke-dasharray: 2 5 2 5;";
        return result;
    }

    protected void outputTextLine(double x, double y, String line, int modifiers) {
        
        Element elt = builder.element("text");
        elt.setAttribute("x", Double.toString(x));
        elt.setAttribute("y", Double.toString(y));
        elt.setAttribute("font-family", getFontFamily());
        elt.setAttribute("font-size", Integer.toString(getFontSize()));
        
        StringBuilder style = new StringBuilder("fill: " + colorToString(getColor()) + "; ");
        if ((getFontStyle() & FONT_BOLD) > 0) style.append("font-weight: bold; ");
        if ((getFontStyle() & FONT_ITALIC) > 0) style.append("font-style: italic; ");
        if ((getFontStyle() & FONT_UNDERLINE) > 0) style.append("text-decoration: underline; ");
        
        elt.setAttribute("style", style.toString());
            
        elt.addContent(builder.text(line));
        rootElement.addContent(elt);
    }

    protected Rectangle2D getTextBounds(String line) {
        return getFont().getStringBounds(line, virtualDevice.getFontRenderContext());
    }
}