/*
 * Created on Jul 15, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements;

import java.util.ArrayList;

import org.jdom.Element;

import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.libraries.scripting.DefaultElementScopeFactory;
import diagnose.graphicengine.libraries.scripting.ElementScope;
import diagnose.graphicengine.libraries.scripting.ElementScopeFactory;
import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XmlConstants;

/**
 * Extends CommonElement with ability to script its methods
 */
public class ActiveElement extends CommonElement
                           implements SubmissiveElement {
    
    private ElementScopeFactory elementScopeFactory = null;
    private ElementScope elementScope = null;

    public ActiveElement() {
        super();    
    }

    public int fixateLastControlPoint(Point p) {
        Integer result = null;
        if (canRunFunction(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT))
            result = (Integer)getElementScope().callFunction(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT, p);
        return (result != null ? result.intValue() : CREATION_CONTINUES);
    }

    public int cancelCreation() {
        Integer result = null;
        if (canRunFunction(XmlConstants.METHOD_CANCEL_CREATION))
            result = (Integer)getElementScope().callFunction(XmlConstants.METHOD_CANCEL_CREATION);
        return (result != null ? result.intValue() : ELEMENT_NOT_VIABLE);
    }

    public ArrayList getControlPointElements() {
        if (canRunFunction(XmlConstants.METHOD_GET_CONTROL_POINT_ELEMENTS))
            return (ArrayList)ElementScope.unwrap(getElementScope().callFunction(XmlConstants.METHOD_GET_CONTROL_POINT_ELEMENTS));
        else return new ArrayList(1);
    }

    public void paint(AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_PAINT)) {
            device.setDefaultValues();
            getElementScope().callFunction(XmlConstants.METHOD_PAINT, device);
        }
    }

    public VisualElement receivesMouseEvent(Point p, AbstractDevice device) {
        Object result = null;
        if (canRunFunction(XmlConstants.METHOD_RECEIVE_MOUSE_EVENT))
            result = getElementScope().callFunction(XmlConstants.METHOD_RECEIVE_MOUSE_EVENT, p, device);
        if (result == null) return null;
        return (VisualElement)ElementScope.unwrap(result);
    }

    public void move(Point direction) {
        if (canRunFunction(XmlConstants.METHOD_MOVE))
            getElementScope().callFunction(XmlConstants.METHOD_MOVE , direction);
    }

    public void scale(Point center, Point coef) {
        if (canRunFunction(XmlConstants.METHOD_SCALE))
            getElementScope().callFunction(XmlConstants.METHOD_SCALE, center, coef);
    }

    public Point snaps(Point p) {
        Object result = null;
        if (canRunFunction(XmlConstants.METHOD_SNAPS))
            result = getElementScope().callFunction(XmlConstants.METHOD_SNAPS, p);
        if (result == null) return null;
        return (Point)ElementScope.unwrap(result);
    }

    public ArrayList getSnapablePoints() {
        Object result = null;
        if (canRunFunction(XmlConstants.METHOD_GET_SNAPABLE_POINTS))
            result = getElementScope().callFunction(XmlConstants.METHOD_GET_SNAPABLE_POINTS);
        if (result == null) return new ArrayList();
        return (ArrayList)ElementScope.unwrap(result);
    }

    public Rect getRectBounds() {
        Object result = null;
        if (canRunFunction(XmlConstants.METHOD_GET_RECT_BOUNDS))
            result = getElementScope().callFunction(XmlConstants.METHOD_GET_RECT_BOUNDS);
        if (result == null) return null;
        return (Rect)ElementScope.unwrap(result);
    }

    public Object clone() {
        ActiveElement result = (ActiveElement) super.clone();
        return result;
    }

    public void deserialize(Element elt, Deserializer deserializer) throws DeserializationException {
        super.deserialize(elt, deserializer);

        if (hasActiveContent()) {
            if (getProperties().hasProperty(XmlConstants.METHOD_PAINT)) {
                getElementScopeFactory().addMethod(XmlConstants.METHOD_PAINT, getProperties().getStringValue(XmlConstants.METHOD_PAINT));
            }
        }
        
    }
    
    public boolean hasActiveContent() {
        if (getProperties(). hasProperty(XmlConstants.PROP_ACTIVE_CONTENT)) 
            return getProperties().getBooleanValue(XmlConstants.PROP_ACTIVE_CONTENT);
        else return false;
    }

    public ElementScope getElementScope() {
        if (elementScope == null) createElementScope();
        return elementScope;
    }

    public ElementScopeFactory getElementScopeFactory() {
        if (elementScopeFactory == null) {
            elementScopeFactory = new DefaultElementScopeFactory(getElementType());
            addMethodIfAny(XmlConstants.METHOD_PAINT);
            addMethodIfAny(XmlConstants.METHOD_MOVE);
            addMethodIfAny(XmlConstants.METHOD_SCALE);
            addMethodIfAny(XmlConstants.METHOD_SNAPS);
            addMethodIfAny(XmlConstants.METHOD_GET_RECT_BOUNDS);
            addMethodIfAny(XmlConstants.METHOD_RECEIVE_MOUSE_EVENT);
            addMethodIfAny(XmlConstants.METHOD_GET_CONTROL_POINT_ELEMENTS);
            addMethodIfAny(XmlConstants.METHOD_GET_SNAPABLE_POINTS);
            addMethodIfAny(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT);
            addMethodIfAny(XmlConstants.METHOD_CANCEL_CREATION);
            addMethodIfAny(XmlConstants.METHOD_ASK_DOMINANT);
            addMethodIfAny(XmlConstants.METHOD_SUGGEST_RELATION);
            addMethodIfAny(XmlConstants.METHOD_CHANGE_SUBMISSIVE);
        }
        return elementScopeFactory;
    }
    
    private void addMethodIfAny(String name) {
        if (getProperties().hasProperty(name))
                elementScopeFactory.addMethod(name, getProperties().getStringValue(name));
    }
    
    /**
     * @return if can run fuction named <code>name</code>
     */
    protected boolean canRunFunction(String name) {
        return (hasActiveContent() && (getProperties().hasProperty(name)));
    }
    
    private void createElementScope() {
        elementScope = getElementScopeFactory().createElementScope();
        elementScope.putObject("properties", getProperties());
        elementScope.putObject("self", this);
        if (getProperties().hasProperty(XmlConstants.PROP_PARENT))
            elementScope.putObject("parent", getProperties().getProperty(XmlConstants.PROP_PARENT).getValue());

        if (getProperties().hasProperty(XmlConstants.METHOD_INIT)) {
            getElementScope().evalString(getProperties().getStringValue(XmlConstants.METHOD_INIT));
        }
    }

    public String getElementType() {
        return XmlConstants.ELEMENT_ACTIVE;
    }

    public void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        if (hasActiveContent() && (getProperties().hasProperty(XmlConstants.METHOD_SUGGEST_RELATION))) {
            ArrayList params = new ArrayList(4);
            params.add(relations);
            params.add(submissiveElt);
            params.add(new Integer(id));
            params.add(position);
            getElementScope().callFunctionEx(XmlConstants.METHOD_SUGGEST_RELATION, params);
        }
    }

    public void changeSubmissive(int id, Point newPosition) {
        if (hasActiveContent()) {
            if (getProperties().hasProperty(XmlConstants.METHOD_CHANGE_SUBMISSIVE)) {
                getElementScope().callFunction(XmlConstants.METHOD_CHANGE_SUBMISSIVE, new Integer(id), newPosition);
            }
        }
    }

    public void askDominant(CommonElement elt) {
        if (hasActiveContent()) {
            if (getProperties().hasProperty(XmlConstants.METHOD_ASK_DOMINANT)) {
                getElementScope().callFunction(XmlConstants.METHOD_ASK_DOMINANT, elt);
            }
        }
    }
    
    
    /**
     * Helper function that accepts suggestion on snapping points
     *
     */
    protected void acceptSuggestionOnSnapPoints(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        ArrayList points = getSnapablePoints();
        for (int i = 0; i < points.size(); i++) {
            Point p = (Point)points.get(i);
            if (p.distance(position) < SNAP_TOLERANCE) {
                relations.add(submissiveElt, id, position);
                break;
            }
        }
    }
}
