/*
 * Created on Feb 5, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements;

import java.util.ArrayList;

import org.jdom.Element;

import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.Lang;
import diagnose.utils.Log;
import diagnose.utils.geometry.Point;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.PointProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XMLSeriazable;
import diagnose.utils.serialization.XmlConstants;

/**
 * Abstract class for all graphic elements forming the document (e.g. lines, 
 * rectangle, etc.)
 * 
 */
public abstract class CommonElement implements VisualElement, 
											XMLSeriazable {
    
    protected Properties properties = new Properties();
    protected Relations relations = new Relations();
    protected PropertyDescriptor propertyDescriptor;
    
    public static final int CREATION_FINISHED = 0;
    public static final int CREATION_CONTINUES = 1;
    public static final int ELEMENT_VIABLE = 0;
    public static final int ELEMENT_NOT_VIABLE = 1;
    
    public static final int OBJECT_CREATING = 1;
    public static final int OBJECT_DONE= 2;
    
    public static final double SNAP_TOLERANCE = 2;
    
    private int objectState = OBJECT_DONE;
    
    private boolean enabledRelation = true;
    
    /**
     * Creates empty element
     *
     */
    public CommonElement() {
        super();
        Properties properties = getProperties();
        if (properties != null) {
            properties.setProperty(new IntProperty(XmlConstants.PROP_CP_COUNT, 0));
            properties.setPropertyDescriptor(getPropertryDescriptor());
        }
    }
    
    public Properties getProperties() {
        return properties;
    }
    
    public void setProperties(Properties properties) {
        this.properties = properties;
    }
    
    /**
     * @return count of control points in element
     * Simplifies access to properties
     */
    public int getControlPointCount() {
        Integer i = (Integer)((IntProperty)getProperties().getProperty(XmlConstants.PROP_CP_COUNT)).getValue(); 
        return i.intValue();
    }
    
    /**
     * @return count of control points in element
     * Simplifies access to properties
     */
    public void setControlPointCount(int cnt) {
        ((IntProperty)getProperties().getProperty(XmlConstants.PROP_CP_COUNT)).setValue(new Integer(cnt));
    }
    
    /**
     * @return control points in element
     * Simplifies access to properties
     */
    public Point getControlPoint(int id) {
        return (Point)((PointProperty)getProperties().getProperty(XmlConstants.PROP_CP + id)).getValue();
    }
    
    public void setControlPoint(int id, Point p) {
        ((PointProperty)getProperties().getProperty(XmlConstants.PROP_CP + id)).setValue(p);
    }
    
    /**
     * Starts creation of new element 
     */
    public void startCreation() {
        objectState = OBJECT_CREATING;
    }
    
    /**
     * Sets position of last control point - used during creation of new element
     */
    public void setLastControlPoint(Point p) {
        int id = getControlPointCount() - 1;
        ((PointProperty)getProperties().getProperty(XmlConstants.PROP_CP + id)).setValue(p);
    }
    
    /**
     * Adds new control point 
     */
    protected void addControlPoint(Point p) {
        int id = getControlPointCount();
        getProperties().setProperty(new PointProperty(XmlConstants.PROP_CP + id, p));
        setControlPointCount(id + 1);
    }
    
    /**
     * Removes last control point 
     */
    protected void removeLastControlPoint() {
        int id = getControlPointCount() - 1;
        getProperties().removeProperty(XmlConstants.PROP_CP + id);
        setControlPointCount(id);
    }
    
    
    /**
     * Sets last control point to specified position - used during creation 
     * of new element. According the returned value are created new control points 
     * @return CREATION_FINISHED if the element creation was finished no other control
     * points are required, or CREATION_CONTINUES - if the element still requires other 
     * control point
     */
    public abstract int fixateLastControlPoint(Point p);
    
    
    /**
     * Sets last control point to specified position - used during creation 
     * of new element. According the returned value are created new control points 
     * @return CREATION_FINISHED if the element creation was finished no other control
     * points are required, or CREATION_CONTINUES - if the element still requires other 
     * control point
     */
    public int setLastControlPointAsFinal(Point p) {
        int state = fixateLastControlPoint(p);
        if (state == CREATION_CONTINUES)
            addControlPoint((Point)p.clone());
        else objectState = OBJECT_DONE;

        return state;
    }
    
    /**
     * Stops creation of an element.
     * @return ELEMENT_VIABLE - if the element has enough control point to be viable
     * otherwise ELEMENT_NOT_VIABLE
     * 
     */
    public abstract int cancelCreation();
    
    /**
     * @return control point elements for this element
     */
    public abstract ArrayList getControlPointElements();
      
    public Element serialize() {
        Element result = builder.element(getElementType());
        Element props = getProperties().serialize();
        if (props != null) result.getChildren().add(props);
        return result;
    }
    
    public void deserialize(Element elt, Deserializer deserializer) throws DeserializationException {
        setProperties((Properties)deserializer.deserialize(elt.getChild(XmlConstants.ELEMENT_PROPERTIES)));
        
        Element propDescElt = elt.getChild(XmlConstants.ELEMENT_PROPERTY_DESCRIPTOR);
        if (propDescElt != null) {
            
            PropertyDescriptor propDesc = (PropertyDescriptor)deserializer.deserialize(propDescElt);
            setPropertyDescriptor(propDesc);
        } else  {
            setPropertyDescriptor(getPropertryDescriptor());
        }
    }
    
    public Object clone() {
        CommonElement result;
        try {
            result = (CommonElement)this.getClass().newInstance();
        } catch (Exception e) {
            e.printStackTrace();
            Log.warn(Lang.get("log.unabletoclone"));
            return null;
        }
        result.setProperties((Properties)this.getProperties().clone());
        
        return result;
    }
    
    public int getObjectState() {
        return objectState;
    }
    
    public void setObjectState(int objectState) {
        this.objectState = objectState;
    }
    
    public PropertyDescriptor getPropertryDescriptor() {
        return propertyDescriptor;
    }
    
    public void setPropertyDescriptor(PropertyDescriptor propertyDescriptor) {
        this.propertyDescriptor = propertyDescriptor;
        getProperties().setPropertyDescriptor(propertyDescriptor);
    }
    
    
    /**
     * Submissive element suggests to the dominant to share point with id
     * In case dominant element accepts this, relations is stored in relations field
     * This method is only for internal purpose and other  elements should call
     * <code>suggestRelation(SubmissiveElement submissiveElt, int id, Point position)</code>
     * instead.
     */
    protected abstract void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position);
    
    /**
     * Submissive element suggests to the dominant to share point with id
     * In case dominant element accepts this, relations is stored in relations field 
     */
    public void suggestRelation(SubmissiveElement submissiveElt, int id, Point position) {
        suggestRelation(getRelations(), submissiveElt, id, position);
    }
    
    public Relations getRelations() {
        return relations;
    }

    public boolean isEnabledRelation() {
        return enabledRelation;
    }

    public void setEnabledRelation(boolean enabledRelation) {
        this.enabledRelation = enabledRelation;
    }
}