/*
 * Created on Feb 5, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import org.jdom.Element;

import diagnose.graphicengine.Manager;
import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.libraries.DependencyException;
import diagnose.graphicengine.libraries.LibElementIdentifier;
import diagnose.graphicengine.libraries.Librarian;
import diagnose.graphicengine.libraries.Library;
import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.Lang;
import diagnose.utils.Log;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.Property;
import diagnose.utils.properties.PropertyReadListener;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XMLSeriazable;
import diagnose.utils.serialization.XmlConstants;

/**
 *
 */
public abstract class ComposedElement extends CommonElement {
    
    protected ArrayList elements = new ArrayList();
    
    public void paint(AbstractDevice device) {
        for (int i = 0; i < elements.size(); i++) {
            getElement(i).paint(device);
        }
    }
    
    public void addElement(CommonElement elt) {
        if (elt.getProperties().getProperty(XmlConstants.PROP_ZINDEX) == null) {
            int zindex = getBiggestZIndex() + 1;
            elt.getProperties().setProperty(new IntProperty(XmlConstants.PROP_ZINDEX, zindex));
        }
        
        elements.add(elt);
        // terrible hack should have own property class
        elt.getProperties().setProperty(new IntProperty(XmlConstants.PROP_PARENT, 1));
        elt.getProperties().getProperty(XmlConstants.PROP_PARENT).setModifiers(Property.VIRTUAL);
        elt.getProperties().getProperty(XmlConstants.PROP_PARENT).setPropertyReadListener(
                new PropertyReadListener() {

                    public Object getValue() {
                        return getSelf();
                    }} );
        reorderElements();
    }
    
    public void removeElement(CommonElement elt) {
        elements.remove(elt);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#receivesMouseEvent(diagnose.graphicengine.utils.Point)
     */
    public VisualElement receivesMouseEvent(Point p, AbstractDevice device) {
        for (int i = elements.size() - 1; i >= 0; i--) {
            CommonElement elt = (CommonElement)elements.get(i); 
            if (elt.receivesMouseEvent(p, device) != null) return elt;
        }
        return null;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#move(diagnose.graphicengine.utils.Point)
     */
    public void move(Point p) {
        for (int i = 0; i < elements.size(); i++) {
            CommonElement elt = getElement(i);
            elt.setEnabledRelation(false);
            elt.move(p);
            elt.setEnabledRelation(true);
        }
        
        if (isEnabledRelation()) 
            getRelations().move(p);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#setLastControlPointAsFinal(diagnose.graphicengine.utils.Point)
     */
    public int fixateLastControlPoint(Point p) {
        setLastControlPoint(p);
        return CREATION_FINISHED;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#cancelCreation()
     */
    public int cancelCreation() {
        return ELEMENT_NOT_VIABLE;
    }

    public Object clone() {
        ComposedElement result;
        try {
            result = (ComposedElement)this.getClass().newInstance();
        } catch (Exception e) {
            Log.error(Lang.get("log.cloneerror"));
            return null;
        }
        
        result.elements.clear();
        for (int i = 0; i < elements.size(); i++) {
            CommonElement elt = (CommonElement)elements.get(i);
            result.addElement((CommonElement)elt.clone());
        }
        result.properties = (Properties)properties.clone();

        return result;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#snaps(diagnose.graphicengine.utils.Point, int)
     */
    public Point snaps(Point p) {
        Point result = null;
        for (int i = 0; i < elements.size(); i++) {
            Point px = ((CommonElement)elements.get(i)).snaps(p);
            if ((px != null) && (result == null)) result = px;
            else if ((px != null) && (result.distance(p) > px.distance(p))) result = px;
        }
        return result;
    }
    
    /**
     * @return count of elements in element
     */
    public int getElementsCount() {
        return elements.size();
    }
    
    /**
     * @return element with id
     */
    public CommonElement getElement(int id) {
        return (CommonElement)elements.get(id);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#getSnapablePoints()
     */
    public ArrayList getSnapablePoints() {
        ArrayList result = new ArrayList();
        for (int i = 0; i < elements.size(); i++)
            result.addAll(getElement(i).getSnapablePoints());
        return result;
    }
    
    /**
     * @return true if element contains specified element (deeply searchs)
     */
    public boolean containsElement(CommonElement searched) {
        for (int i = 0; i < elements.size(); i++) {
            Object elt = elements.get(i); 
            if (elt instanceof ComposedElement)
                return ((ComposedElement)elt).containsElement(searched);
            else
                if (elt == searched) return true;
        }
            
        return false;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#getControlPointElements()
     */
    public ArrayList getControlPointElements() {
        return null;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#scale(diagnose.graphicengine.utils.Point, diagnose.graphicengine.utils.Point)
     */
    public void scale(Point center, Point coef) {        
        for (int i = 0; i < elements.size(); i++) {
            CommonElement elt = getElement(i);
            elt.setEnabledRelation(false);
            elt.scale(center, coef);
            elt.setEnabledRelation(true);
        }
        
        if (isEnabledRelation()) getRelations().scale(center, coef);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#getShapeRect()
     */
    public Rect getRectBounds() {
        ArrayList a = new ArrayList();
        for (int i = 0; i < elements.size(); i++) {
            Rect r = ((VisualElement)elements.get(i)).getRectBounds();
            if (r != null) {
                a.add(r.getPosition());
                a.add(r.getOppositePoint());
            }
        }
        return (a.size() > 0 ? new Rect(a) : null);
    }
        
    public Element serialize() {
        Element result = super.serialize();
        for (int i = 0; i < elements.size(); i++) {
            XMLSeriazable elt = (XMLSeriazable)elements.get(i);
            Element e = elt.serialize();
            if (e != null) {
                result.getChildren().add(e);
            }
        }
        return result;
    }
    
    public void deserialize(Element elt, Deserializer deserializer) throws DeserializationException {
        super.deserialize(elt, deserializer);
        for (int i = 0; i < elt.getChildren().size(); i++) {
            Element e = (Element)elt.getChildren().get(i);
            if ((!e.getName().equals(XmlConstants.ELEMENT_PROPERTIES) && 
                    (!e.getName().equals(XmlConstants.ELEMENT_PROPERTY_DESCRIPTOR)) &&
                    (!e.getName().equals(XmlConstants.ELEMENT_GUIDE_LINES))))
                addElement((CommonElement)deserializer.deserialize(e));
        }
        
        reorderElements();
    }
    
    /**
     * Creates instances of all library elements
     * @return list of elements which weren't found in system
     */
    public ArrayList instatiateLibraryElements(Librarian librarian, boolean positionCorrection) {
        
        ArrayList result = new ArrayList(2);

        for (int i = 0; i < getElementsCount(); i++) {
            if (getElement(i) instanceof ComposedElement)
                result.addAll(((ComposedElement)getElement(i)).instatiateLibraryElements(librarian, positionCorrection));
            
            if (getElement(i) instanceof LibraryElement) {
                LibraryElement elt = (LibraryElement)getElement(i);
                
                String libName = elt.getProperties().getStringValue(XmlConstants.PROP_LIBRARY);
                String eltName = elt.getProperties().getStringValue(XmlConstants.PROP_NAME);
                
                Library lib = librarian.getLibrary(libName);
                if (lib == null) {
                    result.add(new LibElementIdentifier(libName, eltName));
                    continue;
                }
                    
                CommonElement inner = null;
                try {
                    inner = lib.getNewElement(eltName);
                } catch (DependencyException e) { }
                if (inner == null) {
                    result.add(new LibElementIdentifier(libName, eltName));
                    continue;
                }
                
                // move element out side of the universe, to be get back
                // by setting properties, required for group elements to be corretly load
                // using be direct setting of controlpoint positions
                if (positionCorrection) {
                    Point dir = new Point(-Manager.MAXIMAL_PADDING_X * 10, -Manager.MAXIMAL_PADDING_Y * 10);
                    inner.move(dir);
                }
                
                // unifies values of properties in LibElement and InnerObject
                ArrayList propNames = elt.getProperties().getPropertiesNames();
                Collections.sort(propNames, new PropertyNameComparator());
                for (int j = 0; j < propNames.size(); j++) {
                    
                    String propName = (String)propNames.get(j);
                    Property eltProperty = elt.getProperties().getProperty(propName);
                    Property innerProperty = inner.getProperties().getProperty(propName);
                    if (innerProperty == null) inner.getProperties().setProperty(eltProperty);
                    else innerProperty.setValue(eltProperty.getValue(), false);
                }
                
                elt.setProperties(inner.getProperties());
                elt.setInnerObject(inner);
                
            }
        }
        
        return result;
    } 

    /**
     * @return list of inner library elements
     */
    public ArrayList getInnerLibElements() {
        ArrayList result = new ArrayList();
        for (int i = 0; i < getElementsCount(); i++) {
            CommonElement elt = getElement(i);
            
            if (elt instanceof ComposedElement)
                result.addAll(((ComposedElement)elt).getInnerLibElements());
            
            if (elt instanceof LibraryElement)
                result.add(((LibraryElement)elt).getIdentifier());
        }
        
        return result;
    }
    
    /**
     * Returns the biggest used zindex value  
     */
    public int getBiggestZIndex() {
        
        if (elements.size() == 0) return 0;
        int result = Integer.MIN_VALUE;
        for (int i = 0; i < elements.size(); i++) {
            int zindex = getElement(i).getProperties().getIntValue(XmlConstants.PROP_ZINDEX);
            if (zindex > result) result = zindex;
        }
        return result;
    }
    
    /**
     * Returns the least used zindex value  
     */
    public int getLeastZIndex() {
        
        if (elements.size() == 0) return 0;
        int result = Integer.MAX_VALUE;
        for (int i = 0; i < elements.size(); i++) {
            int zindex = getElement(i).getProperties().getIntValue(XmlConstants.PROP_ZINDEX);
            if (zindex < result) result = zindex;
        }
        return result;
    }
    
    
    protected void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        for (int i = 0; i < getElementsCount(); i++)
            getElement(i).suggestRelation(relations, submissiveElt, id, position);
    }

    /**
     * Reorders elements according to their z-index
     */
    public void reorderElements() {
        if (elements.size() <= 1) return;
        Collections.sort(elements, zIndexComparator);
    }
    
    
    private static final class PropertyNameComparator implements Comparator {

        public int compare(Object arg0, Object arg1) {
            String s1 = (String)arg0;
            String s2 = (String)arg1;
            if (s1.startsWith(XmlConstants.PROP_CP) &&
                s2.startsWith(XmlConstants.PROP_CP)) {
                try {
                    int a1 = Integer.valueOf(s1.substring(XmlConstants.PROP_CP.length())).intValue();
                    int a2 = Integer.valueOf(s2.substring(XmlConstants.PROP_CP.length())).intValue();
                    if (a1 == a2) return 0;
                    return (a1 > a2 ? -1 : 1);
                } catch (Exception e) {
                    return s1.compareTo(s2);
                }
                
            }
            return s1.compareTo(s2);
        }
        
    }
    
    private static Comparator zIndexComparator = new ZIndexComparator();
    
    private static final class ZIndexComparator implements Comparator {
        public int compare(Object arg0, Object arg1) {
            if ((!(arg0 instanceof CommonElement)) || (!(arg1 instanceof CommonElement))) return 0;
            
            Properties properties1 = ((CommonElement)arg0).getProperties();
            Properties properties2 = ((CommonElement)arg1).getProperties();
            
            IntProperty zindex1prop = (IntProperty)properties1.getProperty(XmlConstants.PROP_ZINDEX);
            IntProperty zindex2prop = (IntProperty)properties2.getProperty(XmlConstants.PROP_ZINDEX);
            
            int zindex1 = (zindex1prop == null ? -1 : ((Integer)zindex1prop.getValue()).intValue());
            int zindex2 = (zindex2prop == null ? -1 : ((Integer)zindex2prop.getValue()).intValue());

            if (zindex1 < zindex2) return -1;
            if (zindex1 == zindex2) return 0;
            return 1;
        }
    }

    /**
     * Deeply enables/disables acceptance of relations
     */
    public void setEnabledRelation(boolean enabledRelation) {
        super.setEnabledRelation(enabledRelation);
        for (int i = 0; i < getElementsCount(); i++)
            getElement(i).setEnabledRelation(enabledRelation);
    }
    
    private ComposedElement getSelf() {
        return this;
    }
}