/*
 * Created on Jun 2, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements;

import java.util.ArrayList;

import org.jdom.Element;

import diagnose.graphicengine.elements.system.RectangleControlPoints;
import diagnose.graphicengine.relation.Relations;
import diagnose.utils.geometry.Geometry;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.DoubleProperty;
import diagnose.utils.properties.PointProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.Property;
import diagnose.utils.properties.PropertyChangedListener;
import diagnose.utils.properties.PropertyReadListener;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XmlConstants;

/**
 * Abstract class for elements that can be specified by position, width, height
 * Class have to have position, width, height, properties
 */
public class RectBasedElement extends ActiveElement {
    
    public RectBasedElement() {
        super();
        createVirtualProperties(getProperties());
    } 
    
    public void startCreation() {
        super.startCreation();
        addControlPoint(new Point(0, 0));
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#fixateLastControlPoint(diagnose.utils.geometry.Point)
     */
    public int fixateLastControlPoint(Point p) {
        
        if (canRunFunction(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT)) return super.fixateLastControlPoint(p);
        
        setLastControlPoint(p);
        if (getControlPointCount() == 2) {
            Point cp0 = getControlPoint(0);
            Point cp1 = getControlPoint(1);
            if (cp1.x < cp0.x) {
                double tmp = cp1.x;
                cp1.x = cp0.x;
                cp0.x = tmp;
            }
            if (cp1.y < cp0.y) {
                double tmp = cp1.y;
                cp1.y = cp0.y;
                cp0.y = tmp;
            }
            
            createControlPointLimitations(getProperties(), getRelations());
            
            return CREATION_FINISHED;
        }
        return CREATION_CONTINUES;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#cancelCreation()
     */
    public int cancelCreation() {
        
        if (canRunFunction(XmlConstants.METHOD_CANCEL_CREATION)) return super.cancelCreation();
        
        return ELEMENT_NOT_VIABLE;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#getControlPointElements()
     */
    public ArrayList getControlPointElements() {
        
        if (canRunFunction(XmlConstants.METHOD_GET_CONTROL_POINT_ELEMENTS)) return super.getControlPointElements();
        
        ArrayList result = new ArrayList();
        result.add(new RectangleControlPoints.TopLeftControlPointElement(this));
        result.add(new RectangleControlPoints.BottomRightControlPointElement(this));
        result.add(new RectangleControlPoints.TopRightControlPointElement(this));
        result.add(new RectangleControlPoints.BottomLeftControlPointElement(this));
        result.add(new RectangleControlPoints.TopControlPointElement(this));
        result.add(new RectangleControlPoints.BottomControlPointElement(this));
        result.add(new RectangleControlPoints.LeftControlPointElement(this));
        result.add(new RectangleControlPoints.RightControlPointElement(this));
        return result;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#move(diagnose.utils.geometry.Point)
     */
    public void move(Point direction) {
        
        if (canRunFunction(XmlConstants.METHOD_MOVE)) {
            super.move(direction);
            return;
        }
        
        getProperties().getPointValue(XmlConstants.PROP_CP_0).translate(direction);
        getProperties().getPointValue(XmlConstants.PROP_CP_1).translate(direction);
        
        if (isEnabledRelation()) getRelations().move(direction);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#scale(diagnose.utils.geometry.Point, diagnose.utils.geometry.Point)
     */
    public void scale(Point center, Point coef) {
        
        if (canRunFunction(XmlConstants.METHOD_SCALE)) {
            super.scale(center, coef);
            return;
        }
        
        for (int i = 0; i < 2; i++) {
            Point p = (Point)getControlPoint(i).clone();
            Point q = Geometry.scale(p, center, coef);
            getControlPoint(i).x = q.x;
            getControlPoint(i).y = q.y;
        }
        
        if (isEnabledRelation()) getRelations().scale(center, coef);
    }
    
    public void deserialize(Element elt, Deserializer deserializer)
    				throws DeserializationException {
        super.deserialize(elt, deserializer);
        createVirtualProperties(properties);
        createControlPointLimitations(getProperties(), getRelations());
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#getRectBounds()
     */
    public Rect getRectBounds() {
        if (getControlPointCount() < 2) return null;
        return new Rect(getControlPoint(0), getControlPoint(1));
    }
    
    public String getElementType() {
        return XmlConstants.ELEMENT_RECT_BASED;
    }

    protected Point getPosition() {
        return getProperties().getPointValue(XmlConstants.PROP_POSITION);
    }
    
    protected double getWidth() {
        return getProperties().getDoubleValue(XmlConstants.PROP_WIDTH);
    }
    
    protected double getHeight() {
        return getProperties().getDoubleValue(XmlConstants.PROP_HEIGHT);
    }
    
    public Object clone() {
        RectBasedElement result = (RectBasedElement)super.clone();
        createVirtualProperties(result.properties);
        return result;
    }
    
    /**
     * @return set of points in the corners of rectangle
     */
    public ArrayList getCornerPoints() {
        ArrayList result = new ArrayList();
        Point position = (Point)getPosition().clone();
        double width = getWidth();
        double height = getHeight();
        result.add(new Point(position.x, position.y));
        result.add(new Point(position.x + width, position.y));
        result.add(new Point(position.x + width, position.y + height));
        result.add(new Point(position.x, position.y + height));
        return result;
    }
    
    /**
     * @return set of points in the middle of each edge
     */
    public ArrayList getEdgePoints() {
        ArrayList result = new ArrayList();
        Point position = (Point)getPosition().clone();
        double width = getWidth();
        double height = getHeight();
        
        result.add(new Point(position.x + width / 2, position.y));
        result.add(new Point(position.x + width / 2, position.y + height));
        result.add(new Point(position.x, position.y + height / 2));
        result.add(new Point(position.x + width, position.y + height / 2));
        return result;        
    }
    
    /////////////////////// Other useful methods and classes //////////////////////////////
        
    
    /**
     * Adds new properties (position, width, height) that overrides
     * ControlPoint0 and ControlPoint1
     */
    public static void createVirtualProperties(final Properties properties) {        
        
        PointProperty positionProperty = new PointProperty(XmlConstants.PROP_POSITION, null);
        positionProperty.setModifiers(Property.VIRTUAL);
        positionProperty.setPropertyChangedListener(new PropertyChangedListener() {

            public Object propertyChanged(Object oldValue, Object newValue) {
                
                Point oVal = properties.getPointValue(XmlConstants.PROP_CP_0);
                Point nVal = (Point)newValue;
                Point direction = new Point(nVal.x - oVal.x, nVal.y - oVal.y);
                Point oldP1 = properties.getPointValue(XmlConstants.PROP_CP_0);
                Point oldP2 = properties.getPointValue(XmlConstants.PROP_CP_1);
                Point p1 = (Point)oldP1.clone();
                Point p2 = (Point)oldP2.clone();
                p1.translate(direction);
                p2.translate(direction);
                
                // have to be in this way, otherwise limitatarors for controlpoints
                // will do very nasty things
                
                PointProperty cp0 = (PointProperty)properties.getProperty(XmlConstants.PROP_CP_0);
                PointProperty cp1 = (PointProperty)properties.getProperty(XmlConstants.PROP_CP_1);
                
                if (p1.x > properties.getPointValue(XmlConstants.PROP_CP_1).x) {
                    // have to be in this order
                    cp1.setValue(new Point(p2.x, oldP2.y));
                    cp0.setValue(new Point(p1.x, oldP1.y));
                } else {
                    // have to be in this order
                    cp0.setValue(new Point(p1.x, oldP1.y));
                    cp1.setValue(new Point(p2.x, oldP2.y));
                }
                
                oldP1 = properties.getPointValue(XmlConstants.PROP_CP_0);
                oldP2 = properties.getPointValue(XmlConstants.PROP_CP_1);
                
                if (p2.y > properties.getPointValue(XmlConstants.PROP_CP_1).y) {
                    // have to be in this order
                    cp1.setValue(new Point(oldP2.x, p2.y));
                    cp0.setValue(new Point(oldP1.x, p1.y));
                } else {
                    // have to be in this order
                    cp0.setValue(new Point(oldP1.x, p1.y));
                    cp1.setValue(new Point(oldP2.x, p2.y));
                }
                
                return newValue;
            }});
        positionProperty.setPropertyReadListener(new PropertyReadListener() {

            public Object getValue() {
                return properties.getPointValue(XmlConstants.PROP_CP_0);
            }});
        
        DoubleProperty widthProperty = new DoubleProperty(XmlConstants.PROP_WIDTH, 0);
        widthProperty.setModifiers(Property.VIRTUAL);
        widthProperty.setPropertyChangedListener(new PropertyChangedListener() {

            public Object propertyChanged(Object oldValue, Object newValue) {

                double nVal = ((Double)newValue).doubleValue();
                Point p1 = properties.getPointValue(XmlConstants.PROP_CP_0);
                Point p2 = (Point)properties.getPointValue(XmlConstants.PROP_CP_1).clone();
                
                p2.x = p1.x + nVal;
                properties.getProperty(XmlConstants.PROP_CP_1).setValue(p2);
                return newValue;
            }});
        widthProperty.setPropertyReadListener(new PropertyReadListener() {

            public Object getValue() {
                return new Double(
                        properties.getPointValue(XmlConstants.PROP_CP_1).x - 
                        properties.getPointValue(XmlConstants.PROP_CP_0).x);
            }});
        
        DoubleProperty heightProperty = new DoubleProperty(XmlConstants.PROP_HEIGHT, 0);
        heightProperty.setModifiers(Property.VIRTUAL);
        heightProperty.setPropertyChangedListener(new PropertyChangedListener() {

            public Object propertyChanged(Object oldValue, Object newValue) {

                double nVal = ((Double)newValue).doubleValue();
                Point p1 = properties.getPointValue(XmlConstants.PROP_CP_0);
                Point p2 = (Point)properties.getPointValue(XmlConstants.PROP_CP_1).clone();
                p2.y = p1.y + nVal;
                properties.getProperty(XmlConstants.PROP_CP_1).setValue(p2);
                return newValue;
            }});
        heightProperty.setPropertyReadListener(new PropertyReadListener() {

            public Object getValue() {
                return new Double(
                        properties.getPointValue(XmlConstants.PROP_CP_1).y - 
                        properties.getPointValue(XmlConstants.PROP_CP_0).y);
            }});
        
        
        properties.setProperty(positionProperty);
        properties.setProperty(widthProperty);
        properties.setProperty(heightProperty);
    }
    
    /**
     * Creates limitations on control points (ControlPoint0 &lt; ControlPoint1) 
     */
    private static void createControlPointLimitations(final Properties properties, final Relations relations) {
        properties.getProperty(XmlConstants.PROP_CP_0).setPropertyChangedListener(
                new PropertyChangedListener() {

                    public Object propertyChanged(Object oldValue, Object newValue) {
                        Point oldVal = (Point)oldValue;
                        Point cp0 = (Point)newValue;
                        Point cp1 = properties.getPointValue(XmlConstants.PROP_CP_1);
                        Point result = new Point();
                        result.x = (cp0.x < cp1.x ? cp0.x : oldVal.x);
                        result.y = (cp0.y < cp1.y ? cp0.y : oldVal.y);
                        
                        relations.scale(cp1, new Point(
                                (cp1.x - result.x) / (cp1.x - oldVal.x),
                                (cp1.y - result.y) / (cp1.y - oldVal.y)));
                        return result;
                    }});

        properties.getProperty(XmlConstants.PROP_CP_1).setPropertyChangedListener(
                new PropertyChangedListener() {

                    public Object propertyChanged(Object oldValue, Object newValue) {
                        Point oldVal = (Point)oldValue;
                        Point cp0 = properties.getPointValue(XmlConstants.PROP_CP_0);
                        Point cp1 = (Point)newValue;
                        Point result = new Point();
                        result.x = (cp1.x > cp0.x ? cp1.x : oldVal.x);
                        result.y = (cp1.y > cp0.y ? cp1.y : oldVal.y);
                        
                        relations.scale(cp0, new Point(
                                        (result.x - cp0.x) / (oldVal.x - cp0.x),
                                        (result.y - cp0.y) / (oldVal.y - cp0.y)));
                                
                        return result;
                    }});
    }
}
