/*
 * Created on Feb 19, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements.implementations;

import java.util.ArrayList;

import org.jdom.Element;

import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.ComposedElement;
import diagnose.graphicengine.elements.RectBasedElement;
import diagnose.graphicengine.elements.system.RectangleControlPoints;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.PointProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.PropertyChangedListener;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XmlConstants;

/**
 * Element grouping other elements
 */
public class GroupElement extends ComposedElement {
    
    private static PropertyDescriptor defaultPropertyDescriptor = new PropertyDescriptor();
    private int doNotReacts = 0;
    
    static {
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_POSITION, Lang.get("property.position"), Lang.get("category.position"), 10);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_WIDTH, Lang.get("property.width"), Lang.get("category.position"), 11);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_HEIGHT, Lang.get("property.height"), Lang.get("category.position"), 12);
    }
    
    public GroupElement() {
        super();
        getProperties().setProperty(new PointProperty(XmlConstants.PROP_CP_0, new Point()));
        getProperties().setProperty(new PointProperty(XmlConstants.PROP_CP_1, new Point()));
        setPropertyDescriptor(defaultPropertyDescriptor);
        
        setControlPointCount(2);
        RectBasedElement.createVirtualProperties(getProperties());
        setPropertyListeners();
    }
    
    /**
     * Sets property changed listeners
     *
     */
    private void setPropertyListeners() {
        getProperties().getProperty(XmlConstants.PROP_CP_0).setPropertyChangedListener(new PropertyChangedListener() {

            public Object propertyChanged(Object oldValue, Object newValue) {
                if (doNotReacts > 0) return newValue;
                
                calculateControlPoints();
                
                Point p1 = (Point)newValue;
                Point p2 = getProperties().getPointValue(XmlConstants.PROP_CP_1);
                
                Point pOld1 = (Point)oldValue;
                Point pOld2 = p2;
                
                p1.x = (p1.x < p2.x ? p1.x : pOld1.x);
                p1.y = (p1.y < p2.y ? p1.y : pOld1.y);
              
                Point pScale = new Point();
                pScale.x = (p2.x - p1.x) / (pOld2.x - pOld1.x);
                pScale.y = (p2.y - p1.y) / (pOld2.y - pOld1.y);
                
                scale(p2, pScale);

                calculateControlPoints();
                
                return newValue;
            }});
        
        getProperties().getProperty(XmlConstants.PROP_CP_1).setPropertyChangedListener(new PropertyChangedListener() {

            public Object propertyChanged(Object oldValue, Object newValue) {
                if (doNotReacts > 0) return newValue;
                
                calculateControlPoints();
                
                Point p1 = getProperties().getPointValue(XmlConstants.PROP_CP_0);
                Point p2 = (Point)newValue;
                
                Point pOld1 = p1;
                Point pOld2 = (Point)oldValue;
                
                p2.x = (p2.x > p1.x ? p2.x : pOld2.x);
                p2.y = (p2.y > p1.y ? p2.y : pOld2.y);
                
                Point pScale = new Point();
                pScale.x = (p2.x - p1.x) / (pOld2.x - pOld1.x);
                pScale.y = (p2.y - p1.y) / (pOld2.y - pOld1.y);
                
                scale(p1, pScale);
                
                return newValue;
            }});
    }
    
    
    public void addElement(CommonElement elt) {
        super.addElement(elt);
        calculateControlPoints();
    }

    public void move(Point p) {
        super.move(p);
        calculateControlPoints();
    }
    
    public void removeElement(CommonElement elt) {
        super.removeElement(elt);
        calculateControlPoints();
    }
    
    protected GroupElement getParent() {
        return this;
        
    }
  
    public int cancelCreation() {
        return ELEMENT_NOT_VIABLE;
    }
    
    public int fixateLastControlPoint(Point p) {
        
        Point px = (Point)p.clone();
        px.translate(-getProperties().getDoubleValue(XmlConstants.PROP_WIDTH) / 2, 
             -getProperties().getDoubleValue(XmlConstants.PROP_HEIGHT) / 2);
        getProperties().getProperty(XmlConstants.PROP_POSITION).setValue(px);
        return CREATION_FINISHED;
    }
    
    public void setLastControlPoint(Point p) {
        double width = getProperties().getDoubleValue(XmlConstants.PROP_WIDTH);
        double height = getProperties().getDoubleValue(XmlConstants.PROP_HEIGHT);
        Point cp0 = (Point)getControlPoint(0).clone();

        Point p1 = (Point)p.clone();
        p1.translate(-width / 2, -height / 2);
        Point p2 = (Point)p1.clone();
        p2.translate(width, height);
        
        doNotReacts++;
        getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
        getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        doNotReacts--;
        
        Point dir = new Point(p1.x - cp0.x, p1.y - cp0.y);
        move(dir);
    }
    

    protected void calculateControlPoints() {
        ArrayList corners = new ArrayList();
        for (int i = 0; i < getElementsCount(); i++) {
            Rect r = getElement(i).getRectBounds();
            if (r == null) return;
            corners.add(r.getPosition());
            corners.add(r.getOppositePoint());
        }
        Rect r = new Rect(corners);
        doNotReacts++;
        getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(r.getPosition());
        getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(r.getOppositePoint());
        doNotReacts--;
    }
        
    public ArrayList getControlPointElements() {
        
        ArrayList result = new ArrayList();

        result.add(new RectangleControlPoints.TopLeftControlPointElement(this));
        result.add(new RectangleControlPoints.BottomRightControlPointElement(this));
        result.add(new RectangleControlPoints.TopRightControlPointElement(this));
        result.add(new RectangleControlPoints.BottomLeftControlPointElement(this));
        result.add(new RectangleControlPoints.TopControlPointElement(this));
        result.add(new RectangleControlPoints.BottomControlPointElement(this));
        result.add(new RectangleControlPoints.LeftControlPointElement(this));
        result.add(new RectangleControlPoints.RightControlPointElement(this));
        
        return result;
    }

    
    public void paint(AbstractDevice device) {
        if (getControlPointCount() > 1) super.paint(device);
    }

    public String getElementType() {
        return XmlConstants.ELEMENT_GROUP;
    }
    
    public void deserialize(Element elt, Deserializer deserializer)
            throws DeserializationException {
        super.deserialize(elt, deserializer);
        RectBasedElement.createVirtualProperties(getProperties());
        setPropertyListeners();
        calculateControlPoints();
    }

    public void setProperties(Properties properties) {
        super.setProperties(properties);
        RectBasedElement.createVirtualProperties(getProperties());
        setPropertyListeners();
    }
    
    public Object clone() {
        GroupElement result = (GroupElement)super.clone();
        RectBasedElement.createVirtualProperties(result.getProperties());
        result.setPropertyListeners();
        return result;
    }
}