/*
 * Created on Feb 10, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements.implementations;

import java.awt.Color;
import java.util.ArrayList;

import diagnose.graphicengine.Manager;
import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.PointBasedElement;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.DefaultArrays;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Abscissa;
import diagnose.utils.geometry.Geometry;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.ColorProperty;
import diagnose.utils.properties.DoubleProperty;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.properties.SelectPropertyDescription;
import diagnose.utils.serialization.XmlConstants;

/**
 *
 */
public class PolyLineElement extends PointBasedElement {
    
    public PolyLineElement() {
        super();
        getProperties().setProperty(new DoubleProperty(XmlConstants.PROP_LINE_WIDTH, 0.35));
        getProperties().setProperty(new IntProperty(XmlConstants.PROP_LINE_STYLE, AbstractDevice.CLEAR_LINE));
        getProperties().setProperty(new ColorProperty(XmlConstants.PROP_COLOR, Color.BLACK));
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#fixateLastControlPoint(diagnose.graphicengine.utils.Point)
     */
    public int fixateLastControlPoint(Point p) {
        if (canRunFunction(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT)) return super.fixateLastControlPoint(p);
        
        setLastControlPoint(p);
        setPropertyDescriptor(getPropertryDescriptor());
        return CREATION_CONTINUES;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#cancelCreation()
     */
    public int cancelCreation() {
        
        if (canRunFunction(XmlConstants.METHOD_CANCEL_CREATION)) return super.cancelCreation();
        
        if (getControlPointCount() > 1) {
            removeLastControlPoint();
            setPropertyDescriptor(getPropertryDescriptor());
            return ELEMENT_VIABLE;
        }
        
        return ELEMENT_NOT_VIABLE;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#paint(diagnose.graphicengine.AbstractDevice)
     */
    public void paint(AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_PAINT))  {
            super.paint(device);
            return;
        }
        
        if (getControlPointCount() < 2) return;
        device.setDefaultValues();
        device.setColor(getProperties().getColorValue(XmlConstants.PROP_COLOR));
        device.setLineWidth(getProperties().getDoubleValue(XmlConstants.PROP_LINE_WIDTH));
        device.setLineStyle(getProperties().getIntValue(XmlConstants.PROP_LINE_STYLE));
        for (int i = 0; i < getControlPointCount() - 1; i++)
            device.line(getControlPoint(i), getControlPoint(i + 1), AbstractDevice.DEFAULT_DRAW);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#receivesMouseEvent(diagnose.graphicengine.utils.Point)
     */
    public VisualElement receivesMouseEvent(Point p, AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_RECEIVE_MOUSE_EVENT)) return super.receivesMouseEvent(p, device);
        
        for (int i = 0; i < getControlPointCount() - 1; i++)
            if (Geometry.isOnLine(getControlPoint(i), getControlPoint(i + 1), p, Manager.PRECISSNESS / device.getZoom())) return this;
        return null;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#snaps(diagnose.graphicengine.utils.Point, int)
     */
    public Point snaps(Point p) {
        if (canRunFunction(XmlConstants.METHOD_SNAPS)) return super.snaps(p);
        
        Point best = null;
        double bestDist = Double.MAX_VALUE;
        for (int i = 0; i < getControlPointCount() - 1; i++) {
            
            Abscissa line = new Abscissa(getControlPoint(i), getControlPoint(i + 1));
            Point act = line.getNearestPoint(p);
            double actDist = act.distance(p);
            if (actDist < bestDist) {
                best = act;
                bestDist = actDist;
            }
        }
        return best;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#getShapeRect()
     */
    public Rect getRectBounds() {
        if (canRunFunction(XmlConstants.METHOD_GET_RECT_BOUNDS)) return super.getRectBounds();
        
        int cnt = getControlPointCount();
        ArrayList a = new ArrayList(cnt);
        for (int i = 0; i < getControlPointCount(); i++)
            a.add(getControlPoint(i));
          
        return new Rect(a);
    }
    
    public String getElementType() {
        return XmlConstants.ELEMENT_POLYLINE;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#getPropertryDescriptor()
     */
    public PropertyDescriptor getPropertryDescriptor() {
        PropertyDescriptor defaultPropertyDescriptor = new PropertyDescriptor();
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_COLOR, Lang.get("property.color"), Lang.get("category.style"), 1);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_LINE_WIDTH, Lang.get("property.linewidth"), Lang.get("category.style"), 2);
        
        SelectPropertyDescription lineStyleDesc = new SelectPropertyDescription(XmlConstants.PROP_LINE_STYLE, Lang.get("property.linestyle"), Lang.get("category.style"), 3);
        DefaultArrays.fillLineStyles(lineStyleDesc);
        defaultPropertyDescriptor.setDescription(lineStyleDesc);
        fillProperyDescriptorCp(defaultPropertyDescriptor);
        return defaultPropertyDescriptor;
    }
    
    public void askDominant(CommonElement elt) {
        if (canRunFunction(XmlConstants.METHOD_ASK_DOMINANT)) {
            super.askDominant(elt);
            return;
        }
        
        for (int i = 0; i  < getControlPointCount(); i++) {
            if (elt != this) elt.suggestRelation(this, i, getControlPoint(i));
        }
    }

    public void changeSubmissive(int id, Point newPosition) {
        if (canRunFunction(XmlConstants.METHOD_CHANGE_SUBMISSIVE)) {
            super.changeSubmissive(id, newPosition);
            return;
        }
        setControlPoint(id, newPosition);
    }
    
    public void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        if (canRunFunction(XmlConstants.METHOD_SUGGEST_RELATION)) super.suggestRelation(relations, submissiveElt, id, position);
        /*
        else {
            for (int i = 0; i < getControlPointCount(); i++) {
                Point p1 = getControlPoint(i);
                Point p2 = getControlPoint((i + 1) % getControlPointCount());
                Point p = new Point((p1.x + p2.x) / 2, (p1.y + p2.y) / 2);
                if (p.distance(position) < SNAP_TOLERANCE) {
                    relations.add(submissiveElt, id, position);
                }
            }
        }
        */
    }
}
