/*
 * Created on Jul 18, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements.implementations;

import java.awt.Color;
import java.awt.Polygon;
import java.util.ArrayList;

import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.PointBasedElement;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.DefaultArrays;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Abscissa;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;
import diagnose.utils.properties.BoolProperty;
import diagnose.utils.properties.ColorProperty;
import diagnose.utils.properties.DoubleProperty;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.properties.SelectPropertyDescription;
import diagnose.utils.serialization.XmlConstants;

/**
 *
 */
public class PolygonElement extends PointBasedElement {

    public PolygonElement() {
        super();
        getProperties().setProperty(new ColorProperty(XmlConstants.PROP_BACKGROUND, Color.WHITE));
        getProperties().setProperty(new BoolProperty(XmlConstants.PROP_TRANSPARENT, false));
        getProperties().setProperty(new DoubleProperty(XmlConstants.PROP_LINE_WIDTH, 0.35));
        getProperties().setProperty(new IntProperty(XmlConstants.PROP_LINE_STYLE, AbstractDevice.CLEAR_LINE));
        getProperties().setProperty(new ColorProperty(XmlConstants.PROP_COLOR, Color.BLACK));
    }
    
    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#fixateLastControlPoint(diagnose.graphicengine.utils.Point)
     */
    public int fixateLastControlPoint(Point p) {
        if (canRunFunction(XmlConstants.METHOD_FIXATE_LAST_CONTROL_POINT)) return super.fixateLastControlPoint(p);
        
        setLastControlPoint(p);
        setPropertyDescriptor(getPropertryDescriptor());
        return CREATION_CONTINUES;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#cancelCreation()
     */
    public int cancelCreation() {
        
        if (canRunFunction(XmlConstants.METHOD_CANCEL_CREATION)) return super.cancelCreation();
        
        if (getControlPointCount() > 2) {
            removeLastControlPoint();
            setPropertyDescriptor(getPropertryDescriptor());
            return ELEMENT_VIABLE;
        }
        
        return ELEMENT_NOT_VIABLE;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#paint(diagnose.graphicengine.AbstractDevice)
     */
    public void paint(AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_PAINT))  {
            super.paint(device);
            return;
        }
        
        if (getControlPointCount() < 2) return;
        device.setDefaultValues();
        double [] x = new double[getControlPointCount()];
        double [] y = new double[getControlPointCount()];
        for (int i = 0; i < getControlPointCount(); i++) {
            Point p = getControlPoint(i);
            x[i] = p.x;
            y[i] = p.y;
        }
        
        if (!getProperties().getBooleanValue(XmlConstants.PROP_TRANSPARENT)) {
            device.setColor(getProperties().getColorValue(XmlConstants.PROP_BACKGROUND));
            device.fillPolygon(x, y, AbstractDevice.DEFAULT_DRAW);    
        }
        
        device.setLineStyle(getProperties().getIntValue(XmlConstants.PROP_LINE_STYLE));
        device.setLineWidth(getProperties().getDoubleValue(XmlConstants.PROP_LINE_WIDTH));
        device.setColor(getProperties().getColorValue(XmlConstants.PROP_COLOR));
        device.polygon(x, y, AbstractDevice.DEFAULT_DRAW);
        
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#receivesMouseEvent(diagnose.graphicengine.utils.Point)
     */
    public VisualElement receivesMouseEvent(Point p, AbstractDevice device) {
        
        Polygon pol = new Polygon();
        for (int i = 0; i < getControlPointCount(); i++) {
            Point px = getControlPoint(i);
            pol.addPoint((int)px.x, (int)px.y);
        }
        
        return (pol.contains((int)p.x, (int)p.y) ? this : null);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#snaps(diagnose.graphicengine.utils.Point, int)
     */
    public Point snaps(Point p) {
        if (canRunFunction(XmlConstants.METHOD_SNAPS)) return super.snaps(p);
        
        Point best = null;
        double bestDist = Double.MAX_VALUE;
        for (int i = 0; i < getControlPointCount(); i++) {
            
            Abscissa line = new Abscissa(getControlPoint(i), getControlPoint((i + 1) % getControlPointCount()));
            Point act = line.getNearestPoint(p);
            double actDist = act.distance(p);
            if (actDist < bestDist) {
                best = act;
                bestDist = actDist;
            }
        }
        return best;
    }

    public void move(Point direction) {
        super.move(direction);
        if (isEnabledRelation()) getRelations().move(direction);
    }

    public void scale(Point center, Point coef) {
        super.scale(center, coef);
        if (isEnabledRelation()) getRelations().scale(center, coef);
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.VisualElement#getShapeRect()
     */
    public Rect getRectBounds() {
        if (canRunFunction(XmlConstants.METHOD_GET_RECT_BOUNDS)) return super.getRectBounds();
        
        int cnt = getControlPointCount();
        ArrayList a = new ArrayList(cnt);
        for (int i = 0; i < getControlPointCount(); i++)
            a.add(getControlPoint(i));
          
        return new Rect(a);
    }
    
    public String getElementType() {
        return XmlConstants.ELEMENT_POLYGON;
    }

    /* (non-Javadoc)
     * @see diagnose.graphicengine.elements.CommonElement#getPropertryDescriptor()
     */
    public PropertyDescriptor getPropertryDescriptor() {
        PropertyDescriptor defaultPropertyDescriptor = new PropertyDescriptor();
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_COLOR, Lang.get("property.color"), Lang.get("category.style"), 1);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_TRANSPARENT, Lang.get("property.transparent"), Lang.get("category.style"), 21);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_BACKGROUND, Lang.get("property.background"), Lang.get("category.style"), 22);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_LINE_WIDTH, Lang.get("property.linewidth"), Lang.get("category.style"), 23);
        
        SelectPropertyDescription lineStyleDesc = new SelectPropertyDescription(XmlConstants.PROP_LINE_STYLE, Lang.get("property.linestyle"), Lang.get("category.style"), 24);
        DefaultArrays.fillLineStyles(lineStyleDesc);
        defaultPropertyDescriptor.setDescription(lineStyleDesc);
        fillProperyDescriptorCp(defaultPropertyDescriptor);
        return defaultPropertyDescriptor;
    }
    
    public void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        if (canRunFunction(XmlConstants.METHOD_SUGGEST_RELATION)) super.suggestRelation(relations, submissiveElt, id, position);
        else acceptSuggestionOnSnapPoints(relations, submissiveElt, id, position);
    }
}