/*
 * Created on Jun 3, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements.implementations;

import java.awt.Color;
import java.util.ArrayList;

import diagnose.graphicengine.devices.AbstractDevice;
import diagnose.graphicengine.elements.RectBasedElement;
import diagnose.graphicengine.elements.VisualElement;
import diagnose.graphicengine.relation.Relations;
import diagnose.graphicengine.relation.SubmissiveElement;
import diagnose.utils.DPIUtils;
import diagnose.utils.DefaultArrays;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.properties.BoolProperty;
import diagnose.utils.properties.ColorProperty;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.properties.SelectPropertyDescription;
import diagnose.utils.properties.StringProperty;
import diagnose.utils.serialization.XmlConstants;

/**
 *
 */
public class TextElement extends RectBasedElement {
    
    private static PropertyDescriptor defaultPropertyDescriptor = new PropertyDescriptor();
    static {
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_POSITION, Lang.get("property.position"), Lang.get("category.position"), 10);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_WIDTH, Lang.get("property.width"), Lang.get("category.position"), 11);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_HEIGHT, Lang.get("property.height"), Lang.get("category.position"), 12);   
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_TEXT, Lang.get("property.text"), Lang.get("category.text"), 1);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_COLOR, Lang.get("property.color"), Lang.get("category.style"), 20);
        
        SelectPropertyDescription fontDesc = new SelectPropertyDescription(XmlConstants.PROP_FONT_FACE, Lang.get("property.font.face"), Lang.get("category.style"), 21);
        DefaultArrays.fillFontNames(fontDesc);

        defaultPropertyDescriptor.setDescription(fontDesc);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_FONT_SIZE, Lang.get("property.font.size"), Lang.get("category.style"), 22);

        SelectPropertyDescription alignDesc = new SelectPropertyDescription(XmlConstants.PROP_ALIGN, Lang.get("property.align"), Lang.get("category.style"), 23);
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_LEFT), Lang.get("property.align.left"));
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_CENTER), Lang.get("property.align.center"));
        alignDesc.addItem(Integer.toString(AbstractDevice.ALIGN_RIGHT), Lang.get("property.align.right"));
        defaultPropertyDescriptor.setDescription(alignDesc);
        
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_BOLD, Lang.get("property.font.bold"), Lang.get("category.style"), 25);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_ITALIC, Lang.get("property.font.italic"), Lang.get("category.style"), 26);
        defaultPropertyDescriptor.setDescription(XmlConstants.PROP_UNDERLINE, Lang.get("property.font.underline"), Lang.get("category.style"), 27);
    }
    
    public TextElement() {
        super();
        getProperties().setProperty(new ColorProperty(XmlConstants.PROP_COLOR, Color.BLACK));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_TEXT, Lang.get("property.text")));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_FONT_FACE, Lang.get("application.defaultfont")));
        getProperties().setProperty(new IntProperty(XmlConstants.PROP_ALIGN, AbstractDevice.ALIGN_LEFT));
        getProperties().setProperty(new IntProperty(XmlConstants.PROP_FONT_SIZE, 10));
        getProperties().setProperty(new BoolProperty(XmlConstants.PROP_BOLD, false));
        getProperties().setProperty(new BoolProperty(XmlConstants.PROP_ITALIC, false));
        getProperties().setProperty(new BoolProperty(XmlConstants.PROP_UNDERLINE, false));
        setPropertyDescriptor(defaultPropertyDescriptor);
    }

    public void paint(AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_PAINT)) {
            super.paint(device);
            return;
        }
        
        if (getControlPointCount() < 2) return;
        device.setDefaultValues();
        device.setColor(getProperties().getColorValue(XmlConstants.PROP_COLOR));
        if (getObjectState() == OBJECT_CREATING) {
            device.setLineWidth(DPIUtils.pixelsToMilimeters(1 / device.getZoom()));
            device.setLineStyle(AbstractDevice.DASH_LINE);
            device.rect(getRectBounds(), AbstractDevice.DEFAULT_DRAW);
            return;
        }
        
        device.setFontSize(getProperties().getIntValue(XmlConstants.PROP_FONT_SIZE));
        device.setFontFamily(getProperties().getStringValue(XmlConstants.PROP_FONT_FACE));
        
        int style = AbstractDevice.FONT_NORMAL;
        if (getProperties().getBooleanValue(XmlConstants.PROP_BOLD)) style |= AbstractDevice.FONT_BOLD;
        if (getProperties().getBooleanValue(XmlConstants.PROP_ITALIC)) style |= AbstractDevice.FONT_ITALIC;
        if (getProperties().getBooleanValue(XmlConstants.PROP_UNDERLINE)) style |= AbstractDevice.FONT_UNDERLINE;
        
        device.setFontStyle(style);
        
        device.setTextAlign(getProperties().getIntValue(XmlConstants.PROP_ALIGN));
        device.text(getProperties().getStringValue(XmlConstants.PROP_TEXT), getRectBounds(), AbstractDevice.DEFAULT_DRAW);
    }

    public VisualElement receivesMouseEvent(Point p, AbstractDevice device) {
        if (canRunFunction(XmlConstants.METHOD_RECEIVE_MOUSE_EVENT)) return super.receivesMouseEvent(p, device);
        return (getRectBounds().contains(p) ? this : null);
    }

    public Point snaps(Point p) {
        
        if (canRunFunction(XmlConstants.METHOD_SNAPS)) return super.snaps(p);
        
        ArrayList points = getEdgePoints();
        points.addAll(getCornerPoints());
        Point bestPoint = null;
        double bestDist = Double.MAX_VALUE;
        for (int i = 0; i < points.size(); i++) {
            Point actPoint = (Point)points.get(i);
            double actDist = actPoint.distance(p);
            if (actDist < bestDist) {
                bestDist = actDist;
                bestPoint = actPoint;
            }
        }
        return bestPoint;
    }

    public ArrayList getSnapablePoints() {
        
        if (canRunFunction(XmlConstants.METHOD_GET_SNAPABLE_POINTS)) return super.getSnapablePoints();
        
        ArrayList result = new ArrayList();
        result.addAll(getEdgePoints());
        return result;
    }

    public String getElementType() {
        return XmlConstants.ELEMENT_TEXT;
    }

    public void suggestRelation(Relations relations, SubmissiveElement submissiveElt, int id, Point position) {
        if (canRunFunction(XmlConstants.METHOD_SUGGEST_RELATION)) super.suggestRelation(relations, submissiveElt, id, position);
        else acceptSuggestionOnSnapPoints(relations, submissiveElt, id, position);
    }
}
