/*
 * Created on Jul 12, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.elements.system;

import diagnose.graphicengine.elements.CommonElement;
import diagnose.utils.geometry.Point;
import diagnose.utils.serialization.XmlConstants;

/**
 * Holds classes for ControlElements used by rectangular elements
 */
public final class RectangleControlPoints {
 
    /**
     * Returns position and manipulates with top left control point
     *
     */
    public static final class TopLeftControlPointElement extends AbstractControlPointElement {
        
        public TopLeftControlPointElement(CommonElement controls) {
            super(controls);
        }
        
        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            return new Point(p1.x, p1.y);
        }    
        
        public void move(Point direction) {
            
            Point p1 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0).clone();
            p1.translate(direction);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
        }
    }
    
    /**
     * Returns position and manipulates with bottom right control point
     *
     */
    public static final class BottomRightControlPointElement extends AbstractControlPointElement {

        public BottomRightControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            return new Point(p2.x, p2.y);
        }
        
        public void move(Point direction) {
            
            Point p2 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1).clone();
            p2.translate(direction);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        }
    }
    
    /**
     * Returns position and manipulates with top control point
     *
     */
    public static final class TopControlPointElement extends AbstractControlPointElement {

        public TopControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            double width = p2.x - p1.x;
            return new Point(p1.x + width / 2, p1.y);
        }
        
        public void move(Point direction) {
            
            Point p1 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0).clone();
            p1.translate(0, direction.y);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
        }
    }
    
    /**
     * Returns position and manipulates with bottom control point
     *
     */
    public static final class BottomControlPointElement extends AbstractControlPointElement {

        public BottomControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            double width = p2.x - p1.x;
            return new Point(p1.x + width / 2, p2.y);
        }
        
        public void move(Point direction) {
            
            Point p2 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1).clone();
            p2.translate(0, direction.y);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        }
    }
    
    /** 
     * Returns position and manipulates with left control point
     *
     */
    public static final class LeftControlPointElement extends AbstractControlPointElement {

        public LeftControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            double height = p2.y - p1.y;
            return new Point(p1.x, p1.y + height / 2);
        }
        
        public void move(Point direction) {
            
            Point p1 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0).clone();
            p1.translate(direction.x, 0);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
        }
    }

    /** 
     * Returns position and manipulates with right control point
     *
     */
    public static final class RightControlPointElement extends AbstractControlPointElement {

        public RightControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            double height = p2.y - p1.y;
            return new Point(p2.x, p1.y + height / 2);
        }
        
        public void move(Point direction) {
            Point p2 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1).clone();
            p2.translate(direction.x, 0);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        }
    }
    
    /** 
     * Returns position and manipulates with bottom left control point
     *
     */
    public static final class BottomLeftControlPointElement extends AbstractControlPointElement {

        public BottomLeftControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            return new Point(p1.x, p2.y);
        }
        
        public void move(Point direction) {
            
            Point p1 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0).clone();
            Point p2 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1).clone();
            p1.translate(direction.x, 0);
            p2.translate(0, direction.y);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        }
    }
    
    /** 
     * Returns position and manipulates with top right control point
     *
     */
    public static final class TopRightControlPointElement extends AbstractControlPointElement {

        public TopRightControlPointElement(CommonElement controls) {
            super(controls);
        }

        public Point getControlPointPosition() {
            Point p1 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0);
            Point p2 = getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1);
            return new Point(p2.x, p1.y);
        }
        
        public void move(Point direction) {
            
            Point p1 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_0).clone();
            Point p2 = (Point)getControls().getProperties().getPointValue(XmlConstants.PROP_CP_1).clone();
            p1.translate(0, direction.y);
            p2.translate(direction.x, 0);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_0).setValue(p1);
            getControls().getProperties().getProperty(XmlConstants.PROP_CP_1).setValue(p2);
        }
    }
}