/*
 * Created on Apr 21, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.graphicengine.libraries;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.jdom.Element;

import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.elements.ComposedElement;
import diagnose.graphicengine.elements.implementations.GroupElement;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.PropertyDescriptor;
import diagnose.utils.properties.StringProperty;
import diagnose.utils.serialization.DeserializationException;
import diagnose.utils.serialization.Deserializer;
import diagnose.utils.serialization.XMLSeriazable;
import diagnose.utils.serialization.XmlConstants;

/**
 * Holds elements
 */
public class Library implements XMLSeriazable {

    private static PropertyDescriptor propertyDescriptor = new PropertyDescriptor();
    
    private Librarian librarian;
    private Properties properties = new Properties();
    private HashMap elements = new HashMap();
    
    static {
        propertyDescriptor.setDescription(XmlConstants.PROP_NAME, Lang.get("property.libraryname"), Lang.get("category.description"), 1);
        propertyDescriptor.setDescription(XmlConstants.PROP_DESCRIPTION, Lang.get("property.librarydesc"), Lang.get("category.description"), 2);
    }
    
    public Library() {
        super();
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_URL, ""));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_NAME, Lang.get("application.defaultlibraryname")));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_DESCRIPTION, ""));
        getProperties().setPropertyDescriptor(getPropertryDescriptor());
    }    
    
    public Librarian getLibrarian() {
        return librarian;
    }

    public void setLibrarian(Librarian librarian) {
        this.librarian = librarian;
    }
    
    public Library(String name, String description) {
        super();
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_URL, name));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_NAME, name));
        getProperties().setProperty(new StringProperty(XmlConstants.PROP_DESCRIPTION, description));
        getProperties().setPropertyDescriptor(getPropertryDescriptor());
    }
    
    public Properties getProperties() {
        return properties;
    }
    
    public void setProperties(Properties properties) {
        this.properties = properties;
    }
    
    public static PropertyDescriptor getPropertyDescriptor() {
        return propertyDescriptor;
    }
    
    public void removeElement(String name) {
        elements.remove(name);
    }
    
    /**
     * Converts document/Selection into group storeable into library
     */
    public void putElement(String name, ComposedElement doc) {
        
        elements.remove(name);
        
        ComposedElement dc = (ComposedElement)doc.clone();
        GroupElement gelt = new GroupElement();
        for (int i = 0; i < dc.getElementsCount(); i++) {
            gelt.addElement(dc.getElement(i));
        }
        
        Point p = gelt.getRectBounds().getPosition();
        gelt.move(new Point(-p.x, -p.y));
        
        gelt.getProperties().setProperty(new StringProperty(XmlConstants.PROP_NAME, name));

        elements.put(name, gelt);
    }
    
    public String getName() {
        return getProperties().getStringValue(XmlConstants.PROP_NAME);
    }

    public String getElementType() {
        return XmlConstants.ELEMENT_LIBRARY;
    }

    public PropertyDescriptor getPropertryDescriptor() {
        return propertyDescriptor;
    }
    
    public String toString() {
        return getProperties().getStringValue(XmlConstants.PROP_NAME);
    }
    
    /**
     * @return new element created from library template
     */
    public CommonElement getNewElement(String name) throws DependencyException {
        
        CommonElement elt = getElement(name);
        if (elt == null) return null;
        if (elt instanceof ComposedElement) {
        	ArrayList al = ((ComposedElement)elt).instatiateLibraryElements(librarian, false);
            if (al.size() > 0) throw new DependencyException(al);
        }
        
        CommonElement result = (CommonElement)elt.clone();
        result.setObjectState(CommonElement.OBJECT_CREATING);
            
        return result;
    }
    
    public CommonElement getElement(String name) {
        return (CommonElement)elements.get(name);
    }
    
    public Object clone() {
        Library result = new Library();
        result.setProperties((Properties)getProperties().clone());

        Iterator it = elements.keySet().iterator();
        while (it.hasNext()) {
            String eltName = (String)it.next();
            result.addElement(eltName, (CommonElement)elements.get(eltName));
        }
        
        return result;
    }
    
    /**
     * @return list of element's names
     */
    public ArrayList getElementsNames() {
        ArrayList result = new ArrayList();
        Object [] names = elements.keySet().toArray();
        for (int i = 0; i < names.length; i++)
            result.add(names[i]);
        return result;
    }
    
    /**
     * Puts element into library without any conversion
     */
    protected void addElement(String name, CommonElement elt) {
        elements.put(name, elt);
    }
    
    public int getElementsCount() {
        return elements.size();
    }

    /* (non-Javadoc)
     * @see diagnose.utils.serialization.XMLSeriazable#serialize()
     */
    public Element serialize() {

        Iterator it = elements.values().iterator();
        Element elt = builder.element(getElementType());
        Element props = getProperties().serialize();
        elt.getChildren().add(props);
        
        while (it.hasNext()) {
            CommonElement item = (CommonElement)it.next();
            Element e = item.serialize();
            if (e != null) {
                PropertyDescriptor propDesc = item.getPropertryDescriptor();
                if (propDesc != null) {
                    e.addContent(propDesc.serialize());
                }
                elt.addContent(e);
            }
        }
        
        return elt;    
  
    }

    /* (non-Javadoc)
     * @see diagnose.utils.serialization.XMLSeriazable#deserialize(org.jdom.Element, diagnose.utils.serialization.Deserializer)
     */
    public void deserialize(Element elt, Deserializer deserializer) throws DeserializationException {
        setProperties((Properties)deserializer.deserialize(elt.getChild(XmlConstants.ELEMENT_PROPERTIES)));
        getProperties().setPropertyDescriptor(getPropertryDescriptor());
        
        for (int i = 0; i < elt.getChildren().size(); i++) {
            Element e = (Element)elt.getChildren().get(i);
            if ((!e.getName().equals(XmlConstants.ELEMENT_PROPERTIES))
                    && (!e.getName().equals(XmlConstants.ELEMENT_PROPERTY_DESCRIPTOR))) {
                CommonElement celt = (CommonElement)deserializer.deserialize(e);
                addElement(celt.getProperties().getStringValue(XmlConstants.PROP_NAME), celt);
            }
        }
    }
}