/*
 * Created on Aug 3, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.gui;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import diagnose.graphicengine.Manager;
import diagnose.graphicengine.devices.CanvasDevice;
import diagnose.graphicengine.elements.CommonElement;
import diagnose.graphicengine.libraries.DependencyException;
import diagnose.gui.menu.InsertMenuActionFactory;
import diagnose.utils.DPIUtils;
import diagnose.utils.IconCache;
import diagnose.utils.Lang;
import diagnose.utils.geometry.Point;
import diagnose.utils.geometry.Rect;

/**
 * Builds side panel with icons
 */
public final class SidePanelFactory {

    private Manager manager;
    
    private static final int COLOUMNS = 3;
    private static final int ICON_SIZE = 32;
    private static final double ICON_SIZE_MM = DPIUtils.pixelsToMilimeters(ICON_SIZE);
    
    private HashMap cache = new HashMap();
    private JFrame parentFrame;
    
    public SidePanelFactory(Manager manager, JFrame parentFrame) {
        super();
        this.manager = manager;
        this.parentFrame = parentFrame;
    }
    
    public JComponent createPanel() {
        
        JPanel result = new JPanel();
        result.setLayout(new BorderLayout());
        
        final JTabbedPane tabs = new JTabbedPane(JTabbedPane.TOP);
        ArrayList libraries = manager.getLibraries();
        for (int i = 0; i < libraries.size(); i++) {
            String libName = (String)libraries.get(i);
            tabs.addTab(libName, createPanelForLibrary(libName));
        }
        
        final JButton hideShowBtn = new JButton(IconCache.getSystemIcon("hidepanel"));
        Dimension dims = new Dimension(20, 20);
        hideShowBtn.setPreferredSize(dims);
        hideShowBtn.setMaximumSize(dims);
        hideShowBtn.setMinimumSize(dims);
        hideShowBtn.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                boolean isVisible = tabs.isVisible();
                tabs.setVisible(!isVisible);
                hideShowBtn.setIcon(!isVisible ? IconCache.getSystemIcon("hidepanel") : IconCache.getSystemIcon("showpanel"));
            }} );
        
        result.add(tabs, BorderLayout.CENTER);
        result.add(hideShowBtn, BorderLayout.NORTH);
        
        return result;
    }
    
    /**
     * @return new panel with grid of element from library
     */
    private JComponent createPanelForLibrary(String libName) {
        ArrayList elements = manager.getElementsInLibrary(libName);
        return createGrid(libName, elements);
    }
    
    private JComponent createGrid(final String libName, final ArrayList elements) {
        
        final int itemsCount = elements.size();
        final int rows = (int)Math.ceil(itemsCount / (float)COLOUMNS);
        
        TableModel tModel = new AbstractTableModel() {
            
            private static final long serialVersionUID = 3812689127956593769L;

            public int getRowCount() {
                return rows;
            }

            public int getColumnCount() {
                return COLOUMNS;
            }

            public Object getValueAt(int rowIndex, int columnIndex) {
                int itemId = rowIndex * COLOUMNS + columnIndex;
                if (itemId >= itemsCount) return null;
                
                String hash = libName + "@@@" + (String)elements.get(itemId);
                ImageIcon icon = (ImageIcon)cache.get(hash);
                if (icon != null) return icon;
                
                icon = createIcon(libName, (String)elements.get(itemId));
                cache.put(hash, icon);
                return icon;
            }

            public Class getColumnClass(int columnIndex) {
                return ImageIcon.class;
            }

            public boolean isCellEditable(int rowIndex, int columnIndex) {
                return false;
            }            
        };
        
        final JTable result = new JTable(tModel);
        result.setRowHeight(ICON_SIZE + 5);
        result.setRowSelectionAllowed(false);
        result.setFocusable(false);
        result.setTableHeader(null);
        result.addMouseListener(new MouseListener() {

            public void mouseClicked(MouseEvent e) {
                
                int itemId = result.getSelectedRow() * COLOUMNS + result.getSelectedColumn();
                if ((itemId < 0) || (itemId >= itemsCount)) return;
                try {
                    startCreation(libName, (String)elements.get(itemId));
                } catch (DependencyException e1) {
                    InsertMenuActionFactory.unableToInsertElement(
                            parentFrame, libName, 
                            (String)elements.get(itemId), 
                            e1.getDependencies());
                }
            }

            public void mousePressed(MouseEvent e) {
            }

            public void mouseReleased(MouseEvent e) {
            }

            public void mouseEntered(MouseEvent e) {
            }

            public void mouseExited(MouseEvent e) {
            }});
        
        
        JScrollPane pane = new JScrollPane(result);
        pane.setPreferredSize(new Dimension((ICON_SIZE + 10) * COLOUMNS, 0));
        
        return pane;
    }
    
    private void startCreation(String libName, String eltName) throws DependencyException {
        manager.setupCreatingLibraryObject(libName, eltName);
    }
    
    private ImageIcon createIcon(String libName, String eltName) {
        
        SplashScreen.setInfo(Lang.get("splash.loadingelement", libName, eltName));
        
        BufferedImage imgbuf = new BufferedImage(ICON_SIZE, ICON_SIZE, BufferedImage.TYPE_INT_ARGB);
        
        Graphics2D gr = (Graphics2D)imgbuf.getGraphics();
        gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        CanvasDevice device = new CanvasDevice();
        device.setCanvas(gr);
        
        try {
            CommonElement elt = manager.getLibrarian().getLibrary(libName).getNewElement(eltName);
            Rect r = elt.getRectBounds();
            if (r == null) return new ImageIcon(imgbuf);
            double width = r.getWidth();
            double height = r.getHeight();
            double maxSize = Math.max(width, height);
            elt.move(new Point(
                    (maxSize - width) / 2,
                    (maxSize - height) / 2));
            
            double scale = Math.min(ICON_SIZE_MM / width, ICON_SIZE_MM / height);
            gr.scale(scale, scale);
            elt.paint(device);
            
        } catch (DependencyException e) {
            gr.setColor(Color.RED);
            gr.setStroke(new BasicStroke(3));
            gr.drawOval(5, 5, ICON_SIZE - 10, ICON_SIZE - 10);
            gr.translate(ICON_SIZE / 2, ICON_SIZE / 2);
            gr.rotate(Math.PI / 4);
            gr.translate(-ICON_SIZE / 2, -ICON_SIZE / 2);
            gr.drawLine(5, ICON_SIZE / 2, ICON_SIZE - 5, ICON_SIZE / 2);
        }
        
        return new ImageIcon(imgbuf);
    }
}
