/*
 * Created on Apr 21, 2005
 *
 * (c) Petr Krajca <pkrajca@iol.cz>
 */

package diagnose.gui.properties;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import diagnose.utils.GUIUtils;
import diagnose.utils.Lang;
import diagnose.utils.properties.BoolProperty;
import diagnose.utils.properties.ColorProperty;
import diagnose.utils.properties.DoubleProperty;
import diagnose.utils.properties.IntProperty;
import diagnose.utils.properties.PointProperty;
import diagnose.utils.properties.Properties;
import diagnose.utils.properties.Property;
import diagnose.utils.properties.PropertyChangeList;
import diagnose.utils.properties.PropertyChangeLog;
import diagnose.utils.properties.PropertyDescription;
import diagnose.utils.properties.StringProperty;

/**
 * Universal property dialog
 */
public class PropertyDialog extends JDialog {

    private static final long serialVersionUID = 2805137527190484076L;

    private JTabbedPane pageSelector = new JTabbedPane();

    private HashMap pages = new HashMap();

    private Properties properties;

    private PropertyChangeList changes;

    /**
     * @throws java.awt.HeadlessException
     */
    public PropertyDialog(JFrame parent, Properties properties)
            throws HeadlessException {
        super(parent, true);
        initForm(properties);
    }

    public PropertyDialog(JDialog parent, Properties properties)
            throws HeadlessException {
        super(parent, true);
        initForm(properties);
    }

    private void initForm(Properties properties) {
        this.properties = properties;
        this.setModal(true);

        JButton okBtn = new JButton(Lang.get("dialog.button.ok"));
        JButton cancelBtn = new JButton(Lang.get("dialog.button.cancel"));
        JPanel bottomPanel = new JPanel();
        GridLayout layout = new GridLayout(1, 0);
        layout.setHgap(8);
        bottomPanel.setLayout(layout);
        bottomPanel.add(okBtn);
        bottomPanel.add(cancelBtn);

        okBtn.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                changes = saveProperties();
            }
        });

        cancelBtn.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                setVisible(false);
                changes = null;
            }
        });

        createPages();
        setTitle(Lang.get("dialog.caption.properties"));
        setSize(400, 200);
        GUIUtils.centerWindow(this);

        pageSelector.setBorder(BorderFactory.createEmptyBorder(8, 8, 8, 8));
        pageSelector.setAlignmentX(Component.LEFT_ALIGNMENT);
        bottomPanel.setBorder(BorderFactory.createEmptyBorder(8, 8, 8, 8));

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(pageSelector, BorderLayout.CENTER);
        getContentPane().add(bottomPanel, BorderLayout.SOUTH);

        getRootPane().setDefaultButton(okBtn);
        GUIUtils.setEscapeListener(this, new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                setVisible(false);
            }
        });

    }

    public PropertyChangeList getChanges() {
        return changes;
    }

    /**
     * Builds a set of pages with properties
     * 
     */
    protected void createPages() {
        Object[] names = properties.getPropertiesNames().toArray();

        Arrays.sort(names, new Comparator() {

            public int compare(Object arg0, Object arg1) {
                String name1 = (String) arg0;
                String name2 = (String) arg1;

                PropertyDescription desc1 = properties
                        .getPropertyDescription(name1);
                PropertyDescription desc2 = properties
                        .getPropertyDescription(name2);

                int pos1 = (desc1 == null ? -1 : desc1.getPosition());
                int pos2 = (desc2 == null ? -1 : desc2.getPosition());

                if (pos1 < pos2)
                    return -1;
                if (pos1 == pos2)
                    return 0;
                return 1;
            }
        });

        for (int i = 0; i < names.length; i++) {
            String name = (String) names[i];
            PropertyDescription desc = properties.getPropertyDescription(name);
            if (desc != null) {
                if (pages.get(desc.getCategory()) == null)
                    addPage(desc.getCategory());
                PropertyPage pg = (PropertyPage) pages.get(desc.getCategory());
                pg.addProperty(desc, properties.getProperty(name));
            }
        }
    }

    public void runForm() {
        pack();
        setVisible(true);
    }

    public void addPage(String name) {
        PropertyPage pg = new PropertyPage();
        pages.put(name, pg);
        pageSelector.addTab(name, pg.getPanel());
    }

    public void addProperty(String page, PropertyDescription description,
            Property p) {
        PropertyPage pg = (PropertyPage) pages.get(page);
        pg.addProperty(description, p);
    }

    protected PropertyChangeList saveProperties() {

        PropertyChangeList result = new PropertyChangeList();

        ArrayList problems = validateProperties();
        if (problems.size() > 0) {
            StringBuilder buf = new StringBuilder();
            for (int i = 0; i < problems.size(); i++) {
                buf.append(problems.get(i));
                buf.append("\n");
            }
            JOptionPane.showMessageDialog(this, buf.toString(), Lang.get("msg.caption.problem"),
                    JOptionPane.ERROR_MESSAGE);
            return result;
        }

        Iterator i = pages.keySet().iterator();

        while (i.hasNext()) {
            PropertyPage propertyPage = ((PropertyPage) pages.get(i.next()));
            result.add(propertyPage.saveProperties());
        }
        setVisible(false);
        return result;
    }

    /**
     * @return list of all problems in form
     */
    protected ArrayList validateProperties() {
        ArrayList problems = new ArrayList();
        Iterator i = pages.keySet().iterator();

        while (i.hasNext()) {
            PropertyPage propertyPage = ((PropertyPage) pages.get(i.next()));
            problems.addAll(propertyPage.validate());
        }
        return problems;
    }

    protected class PropertyPage {

        private JPanel panel = new JPanel();

        private ArrayList propBoxes = new ArrayList();

        public PropertyPage() {
            super();
            panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
            panel.setBorder(BorderFactory.createEmptyBorder(8, 8, 8, 8));
            panel.setAlignmentX(Component.LEFT_ALIGNMENT);
        }

        public JPanel getPanel() {
            return panel;
        }

        public void addProperty(PropertyDescription description, Property p) {
            PropertyBox propBox = null;
            if (p instanceof PointProperty)
                propBox = new PointPropertyBox(description, p, "invalid.point");

            if (p instanceof BoolProperty)
                propBox = new BoolPropertyBox(description, p, null);

            if (p instanceof IntProperty)
                propBox = new IntPropertyBox(description, p, "invalid.int");

            if (p instanceof DoubleProperty)
                propBox = new DoublePropertyBox(description, p, "invalid.float");

            if (p instanceof ColorProperty)
                propBox = new ColorPropertyBox(description, p, null);

            if (p instanceof StringProperty)
                propBox = new StringPropertyBox(description, p, null, true);

            if (propBox != null) {
                JComponent comp = propBox.getComponent();
                comp.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
                panel.add(comp);
                propBoxes.add(propBox);
            }
        }

        /**
         * @return list all invalid values
         */
        public ArrayList validate() {
            ArrayList problems = new ArrayList();
            for (int i = 0; i < propBoxes.size(); i++) {
                PropertyBox box = (PropertyBox) propBoxes.get(i);
                if (!box.validate())
                    problems.add(Lang.get(box.getInvalidMessage(),
                            box.getDescription().getDescription()));
            }
            return problems;
        }

        /**
         * @return arraylist of PropertyChangeLog informations
         */
        public PropertyChangeList saveProperties() {
            PropertyChangeList result = new PropertyChangeList();
            for (int i = 0; i < propBoxes.size(); i++) {
                PropertyChangeLog changeLog = ((PropertyBox) propBoxes.get(i))
                        .saveValues();
                if (changeLog != null) result.add(changeLog);
            }
            return result;
        }
    }
}
